# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mime implementation.

Mime: Mimicking Centralized Stochastic Algorithms in Federated Learning
    Sai Praneeth Karimireddy, Martin Jaggi, Satyen Kale, Mehryar Mohri,
    Sashank J. Reddi, Sebastian U. Stich, Ananda Theertha Suresh
    https://arxiv.org/abs/2008.03606
"""

from typing import Any, Callable, Mapping, Optional, Sequence, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import models
from fedjax.core import optimizers
from fedjax.core import tree_util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey

import jax
import jax.numpy as jnp

Grads = Params


def create_grads_for_each_client(grad_fn):
  """Builds for_each_client for gradient computation."""

  def client_init(server_params, client_rng):
    client_step_state = {
        'params': server_params,
        'rng': client_rng,
        'num_sum': 0.,
        'grads_sum': jax.tree_util.tree_map(jnp.zeros_like, server_params)
    }
    return client_step_state

  def client_step(client_step_state, batch):
    rng, use_rng = jax.random.split(client_step_state['rng'])
    grads = grad_fn(client_step_state['params'], batch, use_rng)
    num = jnp.sum(batch[client_datasets.EXAMPLE_MASK_KEY])
    grads_sum = tree_util.tree_add(
        tree_util.tree_weight(grads, num), client_step_state['grads_sum'])
    next_client_step_state = {
        'params': client_step_state['params'],
        'rng': rng,
        'num_sum': client_step_state['num_sum'] + num,
        'grads_sum': grads_sum
    }
    return next_client_step_state

  def client_final(server_params, client_step_state):
    del server_params
    client_output = (client_step_state['grads_sum'],
                     client_step_state['num_sum'])
    return client_output

  return for_each_client.for_each_client(client_init, client_step, client_final)


def create_train_for_each_client(grad_fn, base_optimizer):
  """Builds for_each_client for client training."""

  def client_init(shared_input, client_rng):
    client_step_state = {
        'params': shared_input['params'],
        'opt_state': shared_input['opt_state'],
        'rng': client_rng,
        'init_params': shared_input['params'],
        'control_variate': shared_input['control_variate'],
    }
    return client_step_state

  def client_step(client_step_state, batch):
    rng, use_rng = jax.random.split(client_step_state['rng'])
    client_control_variate = grad_fn(client_step_state['init_params'], batch,
                                     use_rng)
    grads = grad_fn(client_step_state['params'], batch, use_rng)
    adjusted_grads = jax.tree_util.tree_multimap(
        lambda g, cc, c: g - cc + c, grads, client_control_variate,
        client_step_state['control_variate'])
    _, params = base_optimizer.apply(adjusted_grads,
                                     client_step_state['opt_state'],
                                     client_step_state['params'])
    next_client_step_state = {
        'params': params,
        'opt_state': client_step_state['opt_state'],
        'rng': rng,
        'init_params': client_step_state['init_params'],
        'control_variate': client_step_state['control_variate'],
    }
    return next_client_step_state

  def client_final(shared_input, client_step_state):
    delta_params = jax.tree_util.tree_multimap(lambda a, b: a - b,
                                               shared_input['params'],
                                               client_step_state['params'])
    return delta_params

  return for_each_client.for_each_client(client_init, client_step, client_final)


@dataclasses.dataclass
class ServerState:
  """State of server passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    opt_state: A pytree representing the base optimizer state.
  """
  params: Params
  opt_state: optimizers.OptState


def mime(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    base_optimizer: optimizers.Optimizer,
    client_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
    grads_batch_hparams: client_datasets.PaddedBatchHParams,
    server_learning_rate: float,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None
) -> federated_algorithm.FederatedAlgorithm:
  """Builds mime.

  Args:
    per_example_loss: A function from (params, batch_example, rng) to a vector
      of loss values for each example in the batch. This is used in both the
      server gradient computation and gradient descent training.
    base_optimizer: Base optimizer to mimic.
    client_batch_hparams: Hyperparameters for batching client dataset for train.
    grads_batch_hparams: Hyperparameters for batching client dataset for server
      gradient computation.
    server_learning_rate: Server learning rate.
    regularizer: Optional regularizer that only depends on params.

  Returns:
    FederatedAlgorithm
  """
  grad_fn = models.grad(per_example_loss, regularizer)
  grads_for_each_client = create_grads_for_each_client(grad_fn)
  train_for_each_client = create_train_for_each_client(grad_fn, base_optimizer)

  def init(params: Params) -> ServerState:
    opt_state = base_optimizer.init(params)
    return ServerState(params, opt_state)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]]
  ) -> Tuple[ServerState, Mapping[federated_data.ClientId, Any]]:
    # Compute full-batch gradient at server params on train data.
    grads_batch_clients = [(cid, cds.padded_batch(grads_batch_hparams), crng)
                           for cid, cds, crng in clients]
    grads_sum_total, num_sum_total = tree_util.tree_sum(
        (co for _, co in grads_for_each_client(server_state.params,
                                               grads_batch_clients)))
    server_grads = tree_util.tree_inverse_weight(grads_sum_total, num_sum_total)
    # Control variant corrected training across clients.
    client_diagnostics = {}
    client_num_examples = {cid: len(cds) for cid, cds, _ in clients}
    batch_clients = [(cid, cds.shuffle_repeat_batch(client_batch_hparams), crng)
                     for cid, cds, crng in clients]
    shared_input = {
        'params': server_state.params,
        'opt_state': server_state.opt_state,
        'control_variate': server_grads
    }
    # Running weighted mean of client updates.
    delta_params_sum = tree_util.tree_zeros_like(server_state.params)
    num_examples_sum = 0.
    for client_id, delta_params in train_for_each_client(
        shared_input, batch_clients):
      num_examples = client_num_examples[client_id]
      delta_params_sum = tree_util.tree_add(
          delta_params_sum, tree_util.tree_weight(delta_params, num_examples))
      num_examples_sum += num_examples
      client_diagnostics[client_id] = {
          'delta_l2_norm': tree_util.tree_l2_norm(delta_params)
      }
    mean_delta_params = tree_util.tree_inverse_weight(delta_params_sum,
                                                      num_examples_sum)

    server_state = server_update(server_state, server_grads, mean_delta_params)
    return server_state, client_diagnostics

  def server_update(server_state, server_grads, mean_delta_params):
    # Server params uses weighted average of client updates, scaled by the
    # server_learning_rate.
    params = jax.tree_util.tree_multimap(
        lambda p, q: p - server_learning_rate * q, server_state.params,
        mean_delta_params)
    opt_state, _ = base_optimizer.apply(server_grads, server_state.opt_state,
                                        server_state.params)
    return ServerState(params, opt_state)

  return federated_algorithm.FederatedAlgorithm(init, apply)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.models.toy_regression."""

from fedjax.legacy.datasets import toy_regression as toy_regression_data
from fedjax.legacy.models import toy_regression as toy_regression_model
import haiku as hk
import tensorflow as tf


class ToyRegressionTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()
    self._rng_seq = hk.PRNGSequence(42)
    self._batch_size = 3
    self._model = toy_regression_model.create_regression_model()
    data, _ = toy_regression_data.load_data(
        num_clients=10, num_domains=2, num_points=100, seed=10)
    self._batch = next(
        data.create_tf_dataset_for_client(data.client_ids[0]).repeat(
            self._batch_size).batch(self._batch_size).as_numpy_iterator())

  def test_backward_pass(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    output = self._model.backward_pass(params, self._batch, next(self._rng_seq))
    self.assertGreaterEqual(output.loss, 0)
    self.assertEqual(output.num_examples, 1)

  def test_evaluate(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    metrics = self._model.evaluate(params, self._batch)
    self.assertContainsSubset(['loss', 'num_examples'], metrics.keys())


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Toy regression data loader."""

import collections
import functools
from typing import Tuple

from fedjax.legacy import core
import numpy as np
import tensorflow as tf
import tensorflow_federated as tff


def domain_id_fn(client_id: str) -> int:
  return int(client_id.split('_')[-1])


def create_toy_regression_data(num_clients: int, num_domains: int,
                               num_points: int, client_domains: np.ndarray,
                               seed: int) -> core.FederatedData:
  r"""Creates toy regression federated data.

  This is a synthetic dataset for a simple regression example for domain
  agnostic learning. Let each domain i, be a set of points x_i_1, ..., x_i_m in
  R. Further, let the average of the points be the center of these points. We
  distribute these points on K clients randomly with the condition that each
  client k can only have points from a single domain i. The goal is to find a
  point that minimizes the maximum distance to all the domain centers.

  Args:
    num_clients: Number of clients to generate.
    num_domains: Number of domains to generate points for.
    num_points: Number of points to generate.
    client_domains: Domain id per client of shape [num_clients].
    seed: Random seed for sampling.

  Returns:
    Federated data containing randomly generated examples of structure:
      x: A numpy array of ones of shape [1].
      y: A numpy array of shape [].
  """
  random_state = np.random.RandomState(seed)

  # Generate points and assign them to domains.
  domains = random_state.randint(0, num_domains, size=[num_points])
  points = random_state.random_sample([num_points]) * 10
  for i in range(num_domains):
    points[domains == i] = points[domains == i] + (i * i)

  # Calculate domain averages and offset points by the center (min + max) / 2.
  avg_point = np.empty(num_domains)
  for i in range(0, num_domains):
    avg_point[i] = np.mean(points[domains == i])
  center = (np.max(avg_point) + np.min(avg_point)) / 2
  points = points - center

  # Group points by domain.
  domain_points = []
  for i in range(num_domains):
    domain_points.append(points[domains == i])
  clients = np.arange(num_clients)

  # Distribute points across clients.
  client_points = [[] for _ in range(num_clients)]
  for i in range(num_domains):
    points_i = domain_points[i]
    clients_i = clients[client_domains == i]
    points_clients = random_state.choice(clients_i, size=len(points_i))

    for j in clients_i:
      client_points[j] = points_i[points_clients == j]

  client_ids = [f'{c}_{d}' for c, d in zip(clients, client_domains)]
  client_data = collections.OrderedDict()
  for c, p in zip(client_ids, client_points):
    client_data[c] = collections.OrderedDict(
        x=tf.expand_dims(np.ones_like(p), axis=-1),
        y=p,
    )

  return tff.simulation.datasets.TestClientData(client_data)


def load_data(num_clients: int, num_domains: int, num_points: int,
              seed: int) -> Tuple[core.FederatedData, core.FederatedData]:
  """Loads toy regression federated data.

  Args:
    num_clients: Number of clients to generate.
    num_domains: Number of domains to generate points for.
    num_points: Number of points to generate.
    seed: Random seed for sampling.

  Returns:
    Train and test toy regression federated data.
  """
  # Randomly assign each client to a domain.
  random_state = np.random.RandomState(seed)
  client_domains = random_state.randint(0, num_domains, size=[num_clients])
  create_data = functools.partial(
      create_toy_regression_data,
      num_clients=num_clients,
      num_domains=num_domains,
      num_points=num_points,
      client_domains=client_domains)
  train_data = create_data(seed=seed + 1)
  test_data = create_data(seed=seed + 2)
  return train_data, test_data



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.agnostic_fed_avg."""

from absl.testing import absltest

from fedjax.algorithms import agnostic_fed_avg
from fedjax.core import client_datasets
from fedjax.core import optimizers

import jax
import jax.numpy as jnp
import numpy.testing as npt


def per_example_loss(params, batch, rng):
  del rng
  return batch['x'] * params['w']


class AgnosticFedAvgTest(absltest.TestCase):

  def test_agnostic_federated_averaging(self):
    algorithm = agnostic_fed_avg.agnostic_federated_averaging(
        per_example_loss=per_example_loss,
        client_optimizer=optimizers.sgd(learning_rate=1.0),
        server_optimizer=optimizers.sgd(learning_rate=0.1),
        client_batch_hparams=client_datasets.ShuffleRepeatBatchHParams(
            batch_size=3, num_epochs=1, seed=0),
        domain_batch_hparams=client_datasets.PaddedBatchHParams(batch_size=3),
        init_domain_weights=[0.1, 0.2, 0.3, 0.4],
        domain_learning_rate=0.01,
        domain_algorithm='eg',
        domain_window_size=2,
        init_domain_window=[1., 2., 3., 4.])

    with self.subTest('init'):
      state = algorithm.init({'w': jnp.array(4.)})
      npt.assert_equal(state.params, {'w': jnp.array(4.)})
      self.assertLen(state.opt_state, 2)
      npt.assert_allclose(state.domain_weights, [0.1, 0.2, 0.3, 0.4])
      npt.assert_allclose(state.domain_window,
                          [[1., 2., 3., 4.], [1., 2., 3., 4.]])

    with self.subTest('apply'):
      clients = [
          (b'cid0',
           client_datasets.ClientDataset({
               'x': jnp.array([1., 2., 4., 3., 6., 1.]),
               'domain_id': jnp.array([1, 0, 0, 0, 2, 2])
           }), jax.random.PRNGKey(0)),
          (b'cid1',
           client_datasets.ClientDataset({
               'x': jnp.array([8., 10., 5.]),
               'domain_id': jnp.array([1, 3, 1])
           }), jax.random.PRNGKey(1)),
      ]
      next_state, client_diagnostics = algorithm.apply(state, clients)
      npt.assert_allclose(next_state.params['w'], 3.5555556)
      npt.assert_allclose(next_state.domain_weights,
                          [0.08702461, 0.18604803, 0.2663479, 0.46057943])
      npt.assert_allclose(next_state.domain_window,
                          [[1., 2., 3., 4.], [3., 3., 2., 1.]])
      npt.assert_allclose(client_diagnostics[b'cid0']['delta_l2_norm'],
                          2.8333335)
      npt.assert_allclose(client_diagnostics[b'cid1']['delta_l2_norm'],
                          7.666667)

    with self.subTest('invalid init_domain_weights'):
      with self.assertRaisesRegex(
          ValueError, 'init_domain_weights must sum to approximately 1.'):
        agnostic_fed_avg.agnostic_federated_averaging(
            per_example_loss=per_example_loss,
            client_optimizer=optimizers.sgd(learning_rate=1.0),
            server_optimizer=optimizers.sgd(learning_rate=1.0),
            client_batch_hparams=client_datasets.ShuffleRepeatBatchHParams(
                batch_size=3),
            domain_batch_hparams=client_datasets.PaddedBatchHParams(
                batch_size=3),
            init_domain_weights=[50., 0., 0., 0.],
            domain_learning_rate=0.5)

    with self.subTest('unequal lengths'):
      with self.assertRaisesRegex(
          ValueError,
          'init_domain_weights and init_domain_window must be equal lengths.'):
        agnostic_fed_avg.agnostic_federated_averaging(
            per_example_loss=per_example_loss,
            client_optimizer=optimizers.sgd(learning_rate=1.0),
            server_optimizer=optimizers.sgd(learning_rate=1.0),
            client_batch_hparams=client_datasets.ShuffleRepeatBatchHParams(
                batch_size=3),
            domain_batch_hparams=client_datasets.PaddedBatchHParams(
                batch_size=3),
            init_domain_weights=[0.1, 0.2, 0.3, 0.4],
            domain_learning_rate=0.5,
            init_domain_window=[1, 2])

  def test_update_domain_weights(self):
    domain_weights = jnp.array([0.1, 0.2, 0.3, 0.4])
    domain_loss = jnp.array([10., 0., 5., 0.])
    domain_learning_rate = 0.1

    with self.subTest('eg'):
      npt.assert_allclose(
          agnostic_fed_avg.update_domain_weights(
              domain_weights,
              domain_loss,
              domain_learning_rate,
              domain_algorithm='eg'),
          [0.198931, 0.14636526, 0.36197326, 0.2927305])

    with self.subTest('none'):
      npt.assert_allclose(
          agnostic_fed_avg.update_domain_weights(
              domain_weights,
              domain_loss,
              domain_learning_rate,
              domain_algorithm='none'), [0.1, 0.2, 0.3, 0.4])

    with self.subTest('INVALID'):
      with self.assertRaisesRegex(ValueError, 'Unsupported domain algorithm'):
        agnostic_fed_avg.update_domain_weights(
            domain_weights,
            domain_loss,
            domain_learning_rate,
            domain_algorithm='INVALID')

  def test_create_domain_metrics_for_each_client(self):
    num_domains = 4
    shared_input = {
        'params': {
            'w': jnp.array(4.)
        },
        'alpha': jnp.array([0.1, 0.2, 0.3, 0.4])
    }
    padded_batch_clients = [
        (b'cid0', [{
            'x': jnp.array([1., 2., 4.]),
            'domain_id': jnp.array([1, 0, 0]),
            '__mask__': jnp.array([True, True, True])
        }, {
            'x': jnp.array([3., 6., 1.]),
            'domain_id': jnp.array([0, 2, 2]),
            '__mask__': jnp.array([True, True, False])
        }], jax.random.PRNGKey(0)),
        (b'cid1', [{
            'x': jnp.array([8., 10., 5.]),
            'domain_id': jnp.array([1, 3, 1]),
            '__mask__': jnp.array([True, True, False])
        }], jax.random.PRNGKey(1)),
    ]
    func = agnostic_fed_avg.create_domain_metrics_for_each_client(
        per_example_loss, num_domains)
    client_domain_metrics = dict(func(shared_input, padded_batch_clients))
    npt.assert_allclose(client_domain_metrics[b'cid0']['beta'], 0.8)
    npt.assert_allclose(client_domain_metrics[b'cid0']['domain_loss'],
                        [36., 4., 24., 0.])
    npt.assert_allclose(client_domain_metrics[b'cid0']['domain_num'],
                        [3., 1., 1., 0.])
    npt.assert_allclose(client_domain_metrics[b'cid1']['beta'], 0.6)
    npt.assert_allclose(client_domain_metrics[b'cid1']['domain_loss'],
                        [0., 32., 0., 40.])
    npt.assert_allclose(client_domain_metrics[b'cid1']['domain_num'],
                        [0., 1., 0., 1.])

  def test_create_scaled_loss(self):
    num_domains = 4
    scaled_loss = agnostic_fed_avg.create_scaled_loss(per_example_loss,
                                                      num_domains)
    params = {'w': jnp.array(4.)}
    batch = {
        'x': jnp.array([1., 2., 4.]),
        'domain_id': jnp.array([1, 0, 0]),
    }
    rng = jax.random.PRNGKey(0)
    alpha = jnp.array([0.1, 0.2, 0.3, 0.4])
    beta = 0.5
    npt.assert_allclose(scaled_loss(params, batch, rng, alpha, beta), 6.4)

  def test_create_train_for_each_client(self):
    num_domains = 4
    shared_input = {
        'params': {
            'w': jnp.array(4.)
        },
        'alpha': jnp.array([0.1, 0.2, 0.3, 0.4])
    }
    batch_clients = [
        (b'cid0', [{
            'x': jnp.array([1., 2., 4.]),
            'domain_id': jnp.array([1, 0, 0]),
        }, {
            'x': jnp.array([3., 6., 1.]),
            'domain_id': jnp.array([0, 2, 2]),
        }], {
            'rng': jax.random.PRNGKey(0),
            'beta': jnp.array(0.5)
        }),
        (b'cid1', [{
            'x': jnp.array([8., 10., 5.]),
            'domain_id': jnp.array([1, 3, 1]),
        }], {
            'rng': jax.random.PRNGKey(1),
            'beta': jnp.array(0.2)
        }),
    ]
    client_optimizer = optimizers.sgd(learning_rate=1.0)
    func = agnostic_fed_avg.create_train_for_each_client(
        per_example_loss, client_optimizer, num_domains)
    client_delta_params = dict(func(shared_input, batch_clients))
    npt.assert_allclose(client_delta_params[b'cid0']['w'], 6.4)
    npt.assert_allclose(client_delta_params[b'cid1']['w'], 33.)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sampling clients for participation in federated training rounds."""

import abc
from typing import Iterator, List, Tuple

from fedjax.core import client_datasets
from fedjax.core import federated_data
from fedjax.core.typing import PRNGKey

import jax
import numpy as np


class ClientSampler(abc.ABC):
  """Interface for client sampler.

  The output of `ClientSampler.sample` should be directly passed as input into
  `FederatedAlgorithm.apply`. In addition to sampling specific clients and their
  datasets, `ClientSampler` also handles generating a seed PRNGKey for each
  client.
  """

  @abc.abstractmethod
  def sample(
      self
  ) -> List[Tuple[federated_data.ClientId, client_datasets.ClientDataset,
                  PRNGKey]]:
    """Samples a subset of clients."""

  @abc.abstractmethod
  def set_round_num(self, round_num: int):
    """Sets round_num for the next sample round."""


class UniformShuffledClientSampler(ClientSampler):
  """Uniformly samples clients using `FederatedData.shuffled_clients`."""

  def __init__(
      self,
      shuffled_clients_iter: Iterator[Tuple[federated_data.ClientId,
                                            client_datasets.ClientDataset]],
      num_clients: int,
      start_round_num: int = 0):
    self._shuffled_clients_iter = shuffled_clients_iter
    self._num_clients = num_clients
    self._round_num = start_round_num
    # In case of restarting from latest checkpoint, we seek the shuffled clients
    # iterator to the previous `round_num - 1` round.
    for _ in range(self._round_num):
      for _ in range(self._num_clients):
        next(self._shuffled_clients_iter)

  def sample(
      self
  ) -> List[Tuple[federated_data.ClientId, client_datasets.ClientDataset,
                  PRNGKey]]:
    clients = []
    client_rngs = jax.random.split(
        jax.random.PRNGKey(self._round_num), self._num_clients)
    for i in range(self._num_clients):
      client_id, client_dataset = next(self._shuffled_clients_iter)
      clients.append((client_id, client_dataset, client_rngs[i]))
    self._round_num += 1
    return clients

  def set_round_num(self, round_num: int):
    raise NotImplementedError(
        "Use UniformGetClientSampler instead if set_round_num() is needed.")


class UniformGetClientSampler(ClientSampler):
  """Uniformly samples clients using `FederatedData.get_clients`."""

  def __init__(self,
               fd: federated_data.FederatedData,
               num_clients: int,
               seed: int,
               start_round_num: int = 0):
    self._federated_data = fd
    self._num_clients = num_clients
    self._seed = seed
    self._client_ids = list(self._federated_data.client_ids())
    self._round_num = start_round_num

  def sample(
      self
  ) -> List[Tuple[federated_data.ClientId, client_datasets.ClientDataset,
                  PRNGKey]]:
    clients = []
    random_state = get_pseudo_random_state(self._seed, self._round_num)
    client_ids = random_state.choice(
        self._client_ids, size=self._num_clients, replace=False)
    client_rngs = jax.random.split(
        jax.random.PRNGKey(self._round_num), self._num_clients)
    for i, (client_id, client_dataset) in enumerate(
        self._federated_data.get_clients(client_ids)):
      clients.append((client_id, client_dataset, client_rngs[i]))
    self._round_num += 1
    return clients

  def set_round_num(self, round_num: int):
    self._round_num = round_num


def get_pseudo_random_state(seed: int, round_num: int) -> np.random.RandomState:
  """Constructs a deterministic numpy random state."""
  #  Settings for a multiplicative linear congruential generator (aka Lehmer
  #  generator) suggested in 'Random Number Generators: Good Ones are Hard to
  # Find' by Park and Miller.
  mlcg_modulus = 2**(31) - 1
  mlcg_multiplier = 16807
  mlcg_start = np.random.RandomState(seed).randint(1, mlcg_modulus - 1)
  return np.random.RandomState(
      pow(mlcg_multiplier, round_num, mlcg_modulus) * mlcg_start % mlcg_modulus)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.models.stackoverflow."""

from typing import Hashable

from absl.testing import absltest

from fedjax.core import tree_util
from fedjax.models import stackoverflow

import jax
import jax.numpy as jnp
import numpy.testing as npt


class StackOverflowModelTest(absltest.TestCase):

  def check_model(self, model):
    params = model.init(jax.random.PRNGKey(0))
    batch = {
        'x': jnp.ones((5, 3), dtype=jnp.int32),
        'y': jnp.ones((5, 3), dtype=jnp.int32)
    }
    with self.subTest('apply_for_train'):
      preds = model.apply_for_train(params, batch)
      self.assertTupleEqual(preds.shape, (5, 3, 10000 + 4))
    with self.subTest('apply_for_eval'):
      preds = model.apply_for_eval(params, batch)
      self.assertTupleEqual(preds.shape, (5, 3, 10000 + 4))
    with self.subTest('train_loss'):
      preds = model.apply_for_train(params, batch)
      train_loss = model.train_loss(batch, preds)
      self.assertTupleEqual(train_loss.shape, (5,))
    with self.subTest('hashable'):
      self.assertIsInstance(model, Hashable)

  def test_create_lstm_model(self):
    model = stackoverflow.create_lstm_model()
    params = model.init(jax.random.PRNGKey(0))
    self.assertEqual(tree_util.tree_size(params), 4050748)
    self.check_model(model)

  def test_create_lstm_model_share_embeddings(self):
    model = stackoverflow.create_lstm_model(share_input_output_embeddings=True)
    params = model.init(jax.random.PRNGKey(0))
    self.assertEqual(tree_util.tree_size(params), 3090364)
    self.check_model(model)

  def test_expected_length_scale_loss(self):
    batch = {'y': jnp.ones((5, 3), dtype=jnp.int32)}
    preds = jnp.ones((5, 3, 10000))
    expected_length = 3.
    model = stackoverflow.create_lstm_model()
    scaled_model = stackoverflow.create_lstm_model(
        expected_length=expected_length)
    train_loss = model.train_loss(batch, preds)
    scaled_train_loss = scaled_model.train_loss(batch, preds)
    npt.assert_array_equal(train_loss * (1 / expected_length),
                           scaled_train_loss)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FedJAX legacy API."""

from fedjax.legacy import algorithms
from fedjax.legacy import datasets
from fedjax.legacy import models
from fedjax.legacy import training
from fedjax.legacy.core import *

# Please do not use symbols in `core` as they are not part of the public API.
try:
  del core  # pylint: disable=undefined-variable
except NameError:
  pass



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for util."""

from absl.testing import absltest
from fedjax.core import util
import jax
import jax.numpy as jnp
import numpy.testing as npt

jax.config.update('jax_debug_nans', True)


class UtilsTest(absltest.TestCase):

  def test_safe_div(self):
    # Safe division by zero.
    npt.assert_array_equal(
        util.safe_div(jnp.array([1, 2, 3]), jnp.array([0, 1, 2])), [0, 2, 1.5])
    # Safe gradient when division by zero.
    grad = jax.grad(lambda xy: jnp.sum(util.safe_div(xy[0], xy[1])))
    npt.assert_array_equal(
        grad(jnp.array([[1, 2, 3], [0, 1, 2]], dtype=jnp.float32)),
        [[0, 1, 0.5], [0, -2, -3 / 4]])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.datasets.stackoverflow.

This file only tests preprocessing functions.
"""

from absl.testing import absltest
from fedjax.datasets import stackoverflow
import numpy as np
import numpy.testing as npt


class StackoverflowTest(absltest.TestCase):

  def test_preprocess_client(self):
    npt.assert_equal(
        stackoverflow.preprocess_client(
            None, {
                'tokens':
                    np.array([b'this is a test', b'may it not fail'],
                             dtype=np.object),
                'creation_date':
                    np.array([b'1970-01-01', b'2021-04-20']),
                'type':
                    np.array([b'answer', b'question'])
            }), {
                'tokens':
                    np.array([b'this is a test', b'may it not fail'],
                             dtype=np.object),
                'domain_id': [1, 0]
            })

  def test_tokenizer(self):
    tokenizer = stackoverflow.StackoverflowTokenizer(
        vocab=['three', 'four', 'five'])

    examples = {
        'tokens':
            np.array([b'three four five', b'five three four five six'],
                     dtype=np.object)
    }

    with self.subTest('padding and truncation'):
      # Truncation only.
      npt.assert_equal(
          tokenizer.as_preprocess_batch(max_length=2)(examples), {
              'x': [[1, 3], [1, 5]],
              'y': [[3, 4], [5, 3]]
          })
      # Padding and truncation.
      npt.assert_equal(
          tokenizer.as_preprocess_batch(max_length=5)(examples), {
              'x': [[1, 3, 4, 5, 0], [1, 5, 3, 4, 5]],
              'y': [[3, 4, 5, 2, 0], [5, 3, 4, 5, 6]]
          })
      # Padding only.
      npt.assert_equal(
          tokenizer.as_preprocess_batch(max_length=7)(examples), {
              'x': [[1, 3, 4, 5, 0, 0, 0], [1, 5, 3, 4, 5, 6, 0]],
              'y': [[3, 4, 5, 2, 0, 0, 0], [5, 3, 4, 5, 6, 2, 0]]
          })

    with self.subTest('domain_id'):
      npt.assert_equal(
          tokenizer.as_preprocess_batch(max_length=2)({
              **examples, 'domain_id': np.array([1, 0])
          }), {
              'x': [[1, 3], [1, 5]],
              'y': [[3, 4], [5, 3]],
              'domain_id': [1, 0]
          })

    with self.subTest('num_oov_buckets'):
      npt.assert_equal(
          stackoverflow.StackoverflowTokenizer(
              vocab=['hello'],
              num_oov_buckets=3).as_preprocess_batch(max_length=5)(examples), {
                  'x': [[1, 5, 5, 4, 0], [1, 4, 5, 5, 4]],
                  'y': [[5, 5, 4, 2, 0], [4, 5, 5, 4, 5]]
              })

    with self.subTest('unused_default_vocab_size'):
      npt.assert_equal(
          stackoverflow.StackoverflowTokenizer(
              vocab=['three', 'four', 'five', 'six'],
              default_vocab_size=1).as_preprocess_batch(max_length=6)(examples),
          {
              'x': [[1, 3, 4, 5, 0, 0], [1, 5, 3, 4, 5, 6]],
              'y': [[3, 4, 5, 2, 0, 0], [5, 3, 4, 5, 6, 2]]
          })


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Microbenchmarks for batch iteration speed of ClientDataset.

Benchmark results on a Xeon E5-1650 v3 on 2021/04/27:
ClientDataset   mode=train      preprocess=False         0.8456923319026828
TF Dataset      mode=train      preprocess=False         5.315499668009579
ClientDataset   mode=train      preprocess=True  0.9447170910425484
TF Dataset      mode=train      preprocess=True  5.322665546089411
ClientDataset   mode=eval       preprocess=False         0.015498528024181724
TF Dataset      mode=eval       preprocess=False         0.8721390531864017
ClientDataset   mode=eval       preprocess=True  0.021079374011605978
TF Dataset      mode=eval       preprocess=True  1.9080286680255085
"""

import timeit

from absl import app
from fedjax.core import client_datasets
import numpy as np
import tensorflow as tf

# pylint: disable=cell-var-from-loop


def main(_):
  for mode in ['train', 'eval']:
    for preprocess in [False, True]:
      assert (bench_client_dataset(preprocess,
                                   mode) == bench_tf_dataset(preprocess, mode))
      print(
          f'ClientDataset\tmode={mode}\tpreprocess={preprocess}\t',
          timeit.timeit(
              lambda: bench_client_dataset(preprocess, mode), number=100))
      print(
          f'TF Dataset\tmode={mode}\tpreprocess={preprocess}\t',
          timeit.timeit(lambda: bench_tf_dataset(preprocess, mode), number=100))


FAKE_MNIST = {
    'pixels': np.random.uniform(size=(1000, 28, 28)),
    'label': np.random.randint(10, size=(1000,))
}


def f(x):
  return {**x, 'binary_label': x['label'] % 2}


def bench_client_dataset(preprocess, mode, batch_size=128, num_steps=100):
  """Benchmarks ClientDataset."""
  preprocessor = client_datasets.NoOpBatchPreprocessor
  if preprocess:
    preprocessor = preprocessor.append(f)
  dataset = client_datasets.ClientDataset(FAKE_MNIST, preprocessor)
  if mode == 'train':
    batches = dataset.shuffle_repeat_batch(
        batch_size=batch_size, num_steps=num_steps)
  else:
    batches = dataset.padded_batch(
        batch_size=batch_size, num_batch_size_buckets=4)
  n = 0
  for _ in batches:
    n += 1
  return n


def bench_tf_dataset(preprocess, mode, batch_size=128, num_steps=100):
  """Benchmarks TF Dataset."""
  shuffle_buffer = 1000  # size of FAKE_MNIST
  dataset = tf.data.Dataset.from_tensor_slices(FAKE_MNIST)
  if mode == 'train':
    dataset = dataset.shuffle(shuffle_buffer).repeat()
  dataset = dataset.batch(batch_size)
  if mode == 'train':
    dataset = dataset.take(num_steps)
  if preprocess:
    dataset = dataset.map(f)
  n = 0
  for _ in dataset.as_numpy_iterator():
    n += 1
  return n


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface definitions for aggregators."""

from typing import Any, Callable, Iterable, Tuple
from fedjax.core import dataclasses
from fedjax.core import tree_util
from fedjax.core.federated_data import ClientId
from fedjax.core.typing import Params

PyTree = Any
AggregatorState = PyTree


@dataclasses.dataclass
class Aggregator:
  """Interface for algorithms to aggregate.

  This interface defines aggregator algorithms that are used at each round.
  Aggregator state contains any round specific parameters (e.g. number of bits)
  that will be passed from round to round. This state is initialized by
  `init` and passed as input into and returned as output from `aggregate`.
  We strongly recommend using fedjax.dataclass to define state as this provides
  immutability, type hinting, and works by default with JAX transformations.

  The expected usage of Aggregator is as follows:
  ```
   aggregator = mean_aggregator()
   state = aggregator.init()
   for i in range(num_rounds):
     clients_params_and_weights = compute_client_outputs(i)
     aggregated_params, state = aggregator.apply(clients_params_and_weights,
                                                 state)

  ```

  Attributes:
    init: Returns initial state of aggregator.
    apply: Returns the new aggregator state and aggregated params.
  """
  init: Callable[[], AggregatorState]
  apply: Callable[
      [Iterable[Tuple[ClientId, Params, float]], AggregatorState],
      Tuple[Params, AggregatorState]]


@dataclasses.dataclass
class MeanAggregatorState:
  """Mean aggregator is stateless."""


def mean_aggregator() -> Aggregator:
  """Builds (weighted) mean aggregator."""

  def init():
    return MeanAggregatorState()

  def apply(clients_params_and_weights, state):
    def extract_params_and_weight(clients_params_and_weight):
      _, param, weight = clients_params_and_weight
      return param, weight
    params_and_weights = map(extract_params_and_weight,
                             clients_params_and_weights)
    return tree_util.tree_mean(params_and_weights), state

  return Aggregator(init, apply)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EMNIST models."""

from fedjax.core import metrics
from fedjax.core import models

import haiku as hk
import jax
from jax.experimental import stax
import jax.numpy as jnp
import numpy as np


class Dropout(hk.Module):
  """Dropout haiku module."""

  def __init__(self, rate: float = 0.5):
    """Initializes dropout module.

    Args:
      rate: Probability that each element of x is discarded. Must be in [0, 1).
    """
    super().__init__()
    self._rate = rate

  def __call__(self, x: jnp.ndarray, is_train: bool):
    if is_train:
      return hk.dropout(rng=hk.next_rng_key(), rate=self._rate, x=x)
    return x


class ConvDropoutModule(hk.Module):
  """Custom haiku module for CNN with dropout.

  This must be defined as a custom hk.Module because only a single positional
  argument is allowed when using hk.Sequential.
  """

  def __init__(self, num_classes):
    super().__init__()
    self._num_classes = num_classes

  def __call__(self, x: jnp.ndarray, is_train: bool):
    x = hk.Conv2D(output_channels=32, kernel_shape=(3, 3), padding='VALID')(x)
    x = jax.nn.relu(x)
    x = hk.Conv2D(output_channels=64, kernel_shape=(3, 3), padding='VALID')(x)
    x = jax.nn.relu(x)
    x = (
        hk.MaxPool(
            window_shape=(1, 2, 2, 1), strides=(1, 2, 2, 1),
            padding='VALID')(x))
    x = Dropout(rate=0.25)(x, is_train)
    x = hk.Flatten()(x)
    x = hk.Linear(128)(x)
    x = jax.nn.relu(x)
    x = Dropout(rate=0.5)(x, is_train)
    x = hk.Linear(self._num_classes)(x)
    return x


# Defines the expected structure of input batches to the model. This is used to
# determine the model parameter shapes.
_HAIKU_SAMPLE_BATCH = {
    'x': np.zeros((1, 28, 28, 1), dtype=np.float32),
    'y': np.zeros(1, dtype=np.float32)
}
_STAX_SAMPLE_SHAPE = (-1, 28, 28, 1)

_TRAIN_LOSS = lambda b, p: metrics.unreduced_cross_entropy_loss(b['y'], p)
_EVAL_METRICS = {
    'loss': metrics.CrossEntropyLoss(),
    'accuracy': metrics.Accuracy()
}


def create_conv_model(only_digits: bool = False) -> models.Model:
  """Creates EMNIST CNN model with dropout with haiku.

  Matches the model used in:

  Adaptive Federated Optimization
    Sashank Reddi, Zachary Charles, Manzil Zaheer, Zachary Garrett, Keith Rush,
    Jakub Konečný, Sanjiv Kumar, H. Brendan McMahan.
    https://arxiv.org/abs/2003.00295

  Args:
    only_digits: Whether to use only digit classes [0-9] or include lower and
      upper case characters for a total of 62 classes.

  Returns:
    Model
  """
  num_classes = 10 if only_digits else 62

  def forward_pass(batch, is_train=True):
    return ConvDropoutModule(num_classes)(batch['x'], is_train)

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS,
      # is_train determines whether to apply dropout or not.
      train_kwargs={'is_train': True},
      eval_kwargs={'is_train': False})


def create_dense_model(only_digits: bool = False,
                       hidden_units: int = 200) -> models.Model:
  """Creates EMNIST dense net with haiku."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(hidden_units),
        jax.nn.relu,
        hk.Linear(hidden_units),
        jax.nn.relu,
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)


def create_logistic_model(only_digits: bool = False) -> models.Model:
  """Creates EMNIST logistic model with haiku."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)


def create_stax_dense_model(only_digits: bool = False,
                            hidden_units: int = 200) -> models.Model:
  """Creates EMNIST dense net with stax."""
  num_classes = 10 if only_digits else 62
  stax_init, stax_apply = stax.serial(stax.Flatten,
                                      stax.Dense(hidden_units), stax.Relu,
                                      stax.Dense(hidden_units), stax.Relu,
                                      stax.Dense(num_classes))
  return models.create_model_from_stax(
      stax_init=stax_init,
      stax_apply=stax_apply,
      sample_shape=_STAX_SAMPLE_SHAPE,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Basic federated averaging implementation using fedjax.experimental.

This is the basic implementation in that sense that we write out the native
Python for loops over clients by hand. While this is much more straightforward
to implement, it's not the fastest. This is a good starting point for
implementing custom algorithms since it tends to follow pseudocode closely.

Based on the paper:

Communication-Efficient Learning of Deep Networks from Decentralized Data
    H. Brendan McMahan, Eider Moore, Daniel Ramage,
    Seth Hampson, Blaise Aguera y Arcas. AISTATS 2017.
    https://arxiv.org/abs/1602.05629
"""

from typing import Any, Callable, Mapping, Sequence, Tuple

import fedjax

import jax

ClientId = bytes
Grads = fedjax.Params


@fedjax.dataclass
class ServerState:
  """State of server passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    opt_state: A pytree representing the server optimizer state.
  """
  params: fedjax.Params
  opt_state: fedjax.OptState


def federated_averaging(
    grad_fn: Callable[[fedjax.Params, fedjax.BatchExample, fedjax.PRNGKey],
                      Grads],
    client_optimizer: fedjax.Optimizer,
    server_optimizer: fedjax.Optimizer,
    client_batch_hparams: fedjax.ShuffleRepeatBatchHParams
) -> fedjax.FederatedAlgorithm:
  """Builds the basic implementation of federated averaging."""

  def init(params: fedjax.Params) -> ServerState:
    opt_state = server_optimizer.init(params)
    return ServerState(params, opt_state)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[ClientId, fedjax.ClientDataset, fedjax.PRNGKey]]
  ) -> Tuple[ServerState, Mapping[ClientId, Any]]:
    client_diagnostics = {}
    # We use a list here for clarity, but we strongly recommend avoiding loading
    # all client outputs into memory since the outputs can be quite large
    # depending on the size of the model.
    client_delta_params_weights = []
    for client_id, client_dataset, client_rng in clients:
      delta_params = client_update(server_state.params, client_dataset,
                                   client_rng)
      client_delta_params_weights.append((delta_params, len(client_dataset)))
      # We record the l2 norm of client updates as an example, but it is not
      # required for the algorithm.
      client_diagnostics[client_id] = {
          'delta_l2_norm': fedjax.tree_util.tree_l2_norm(delta_params)
      }
    mean_delta_params = fedjax.tree_util.tree_mean(client_delta_params_weights)
    server_state = server_update(server_state, mean_delta_params)
    return server_state, client_diagnostics

  def client_update(server_params, client_dataset, client_rng):
    params = server_params
    opt_state = client_optimizer.init(params)
    for batch in client_dataset.shuffle_repeat_batch(client_batch_hparams):
      client_rng, use_rng = jax.random.split(client_rng)
      grads = grad_fn(params, batch, use_rng)
      opt_state, params = client_optimizer.apply(grads, opt_state, params)
    delta_params = jax.tree_util.tree_multimap(lambda a, b: a - b,
                                               server_params, params)
    return delta_params

  def server_update(server_state, mean_delta_params):
    opt_state, params = server_optimizer.apply(mean_delta_params,
                                               server_state.opt_state,
                                               server_state.params)
    return ServerState(params, opt_state)

  return fedjax.FederatedAlgorithm(init, apply)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.aggregators.aggregator."""
from absl.testing import absltest

from fedjax.aggregators import aggregator
import jax.numpy as jnp
import numpy.testing as npt


class AggregatorTest(absltest.TestCase):

  def test_mean_aggregator(self):
    delta_params_and_weights = [('a', {
        'w': jnp.array([1., 2., 3.])
    }, 2.), ('b', {
        'w': jnp.array([2., 4., 6.])
    }, 4.), ('c', {
        'w': jnp.array([1., 3., 5.])
    }, 2.)]
    mean_aggregator = aggregator.mean_aggregator()

    aggregator_state = mean_aggregator.init()
    mean_params, _ = mean_aggregator.apply(delta_params_and_weights,
                                           aggregator_state)

    npt.assert_array_equal(mean_params['w'], [1.5, 3.25, 5.])


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.hyp_cluster."""

from fedjax.legacy import core
from fedjax.legacy.algorithms import hyp_cluster
import tensorflow as tf


class HypClusterTest(tf.test.TestCase):

  def test_run(self):
    num_classes = 10
    num_clusters = 3
    federated_data, model = core.test_util.create_toy_example(
        num_clients=10,
        num_clusters=num_clusters,
        num_classes=num_classes,
        num_examples=5,
        seed=0)
    rng_seq = core.PRNGSequence(0)
    algorithm = hyp_cluster.HypCluster(
        federated_data=federated_data,
        model=model,
        client_optimizer=core.get_optimizer(
            core.OptimizerName.SGD, learning_rate=0.1),
        server_optimizer=core.get_optimizer(
            core.OptimizerName.SGD, learning_rate=1.0),
        hparams=hyp_cluster.HypClusterHParams(
            train_data_hparams=core.ClientDataHParams(batch_size=5),
            num_clusters=num_clusters),
        rng_seq=rng_seq,
    )

    state = algorithm.init_state()
    for _ in range(10):
      state = algorithm.run_round(state, federated_data.client_ids)

    with self.subTest('num_clusters'):
      self.assertLen(state.cluster_params, num_clusters)

    with self.subTest('maximization'):
      data_hparams = core.ClientDataHParams(batch_size=5)
      cluster_client_ids = hyp_cluster.maximization(federated_data,
                                                    federated_data.client_ids,
                                                    model, state.cluster_params,
                                                    data_hparams)
      for cluster_id, client_ids in enumerate(cluster_client_ids):
        for client_id in client_ids:
          dataset = federated_data.create_tf_dataset_for_client(client_id)
          dataset = core.preprocess_tf_dataset(dataset, data_hparams)
          cluster_loss = []
          for params in state.cluster_params:
            cluster_loss.append(
                core.evaluate_single_client(dataset, model, params)['loss'])
          # Cluster should be the best params for a client because clients are
          # clustered based on empirical loss.
          self.assertEqual(cluster_id, cluster_loss.index(min(cluster_loss)))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple download manager."""

import math
import os.path
import sys
import time
from typing import Callable, Iterator, Optional
import urllib.parse

import requests


def progress(n: int) -> Iterator[int]:
  """A simple generator for tracking progress."""
  start = time.time()
  last_log = -1
  for i in range(n):
    yield i
    elapsed = time.time() - start
    if elapsed - last_log >= 1:
      last_log = elapsed
      eta = elapsed / (i + 1) * (n - i - 1)
      log(f'\r{100*(i + 1)/n:3.0f}%, ETA: {format_duration(eta)}', end='')
  log(f'\r100%, elapsed: {format_duration(time.time() - start)}')


def maybe_download(url: str,
                   cache_dir: Optional[str] = None,
                   progress_: Callable[[int], Iterator[int]] = progress) -> str:
  """Downloads `url` to local disk.

  Args:
    url: URL to download from.
    cache_dir: Where to cache the file. If None, uses default_cache_dir().
    progress_: A callable that yields like range(n), for tracking progress.

  Returns:
    Path to local file.
  """
  # TODO(wuke): Avoid race conditions when downloading the same file from
  # different threads/processes at the same time.
  if cache_dir is None:
    cache_dir = default_cache_dir()
  os.makedirs(cache_dir, exist_ok=True)
  path = os.path.join(cache_dir,
                      os.path.basename(urllib.parse.urlparse(url).path))
  if os.path.exists(path):
    log(f'Reusing cached file {path!r}')
  else:
    log(f'Downloading {url!r} to {path!r}')
    with open(path + '.partial', 'wb') as fo:
      r = requests.get(url, stream=True)
      r.raise_for_status()
      length = int(r.headers['content-length'])
      block_size = 1 << 18
      for _ in progress_((length + block_size - 1) // block_size):
        fo.write(r.raw.read(block_size))
    os.rename(path + '.partial', path)
  return path


def format_duration(seconds: float) -> str:
  """Formats duration in seconds into hours/minutes/seconds."""
  if seconds < 60:
    return f'{seconds:.0f}s'
  elif seconds < 3600:
    minutes = math.floor(seconds / 60)
    seconds -= minutes * 60
    return f'{minutes}m{seconds:.0f}s'
  else:
    hours = math.floor(seconds / 3600)
    seconds -= hours * 3600
    minutes = math.floor(seconds / 60)
    seconds -= minutes * 60
    return f'{hours}h{minutes}m{seconds:.0f}s'


def log(*args, **kwargs):
  print(*args, flush=True, file=sys.stderr, **kwargs)


def default_cache_dir() -> str:
  """Returns default local file cache directory."""
  running_on_colab = 'google.colab' in sys.modules
  if running_on_colab:
    base_dir = '/tmp'
  else:
    base_dir = os.path.expanduser('~')
  cache_dir = os.path.join(base_dir, '.cache/fedjax')
  return cache_dir



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.serialization."""

from absl.testing import absltest
from fedjax.core import serialization
import numpy as np
import numpy.testing as npt


class SerializationTest(absltest.TestCase):

  def test_dict(self):
    original = {
        'int32':
            np.arange(4, dtype=np.int32).reshape([2, 2]),
        'float64':
            -np.arange(4, dtype=np.float64).reshape([1, 4]),
        'bytes':
            np.array([b'a', b'bc', b'def'], dtype=np.object).reshape([3, 1]),
    }
    output = serialization.msgpack_deserialize(
        serialization.msgpack_serialize(original))
    self.assertCountEqual(output, original)
    self.assertEqual(output['int32'].dtype, np.int32)
    npt.assert_array_equal(output['int32'], original['int32'])
    self.assertEqual(output['float64'].dtype, np.float64)
    npt.assert_array_equal(output['float64'], original['float64'])
    self.assertEqual(output['bytes'].dtype, np.object)
    npt.assert_array_equal(output['bytes'], original['bytes'])

  def test_nested_list(self):
    original = [
        np.arange(4, dtype=np.int32).reshape([2, 2]),
        [
            -np.arange(4, dtype=np.float64).reshape([1, 4]),
            [
                np.array([b'a', b'bc', b'def'],
                         dtype=np.object).reshape([3, 1]), []
            ]
        ]
    ]
    output = serialization.msgpack_deserialize(
        serialization.msgpack_serialize(original))
    int32_array, rest = output
    self.assertEqual(int32_array.dtype, np.int32)
    npt.assert_array_equal(int32_array, original[0])
    float64_array, rest = rest
    self.assertEqual(float64_array.dtype, np.float64)
    npt.assert_array_equal(float64_array, original[1][0])
    bytes_array, rest = rest
    self.assertEqual(bytes_array.dtype, np.object)
    npt.assert_array_equal(bytes_array, original[1][1][0])
    self.assertEqual(rest, [])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""InMemoryFederatedData for small custom datasets."""

from typing import Callable, Iterable, Iterator, Mapping, Optional, Tuple

from fedjax.core import client_datasets
from fedjax.core import federated_data
import numpy as np


class InMemoryFederatedData(federated_data.FederatedData):
  """A simple wrapper over a concrete fedjax.FederatedData for small in memory datasets.

  This is useful when we wish to create a smaller FederatedData that fits in
  memory. Here is a simple example to create a fedjax.InMemoryFederatedData, ::

    client_a_data = {
        'x': np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
        'y': np.array([7, 8])
    }
    client_b_data = {'x': np.array([[9.0, 10.0, 11.0]]), 'y': np.array([12])}
    client_to_data_mapping = {'a': client_a_data, 'b': client_b_data}

    fedjax.InMemoryFederatedData(client_to_data_mapping)

  Returns:
    A fedjax.InMemoryDataset corresponding to client_to_data_mapping.
  """

  def __init__(self,
               client_to_data_mapping: Mapping[federated_data.ClientId,
                                               Mapping[str, np.ndarray]],
               preprocess_client: federated_data
               .ClientPreprocessor = federated_data.NoOpClientPreprocessor,
               preprocess_batch: client_datasets
               .BatchPreprocessor = client_datasets.NoOpBatchPreprocessor):
    """Initializes the in memory federated dataset.

      Data of each client is a mapping from feature names to numpy arrays. For
      example, for emnist image classification,
      {'x': X, 'y': y}, where X is a matrix of shape (num_data_points, 28, 28)
      and y is a matrix of shape (num_data_points).

    Args:
      client_to_data_mapping: A mapping from client_id to data of each client.
      preprocess_client: federated_data.ClientPreprocessor to preprocess each
        client data.
      preprocess_batch: client_datasets.BatchPreprocessor to preprocess batch of
        data.
    """
    self._preprocess_client = preprocess_client
    self._preprocess_batch = preprocess_batch
    self._client_to_data_mapping = client_to_data_mapping
    self._client_ids = sorted(self._client_to_data_mapping.keys())
    self._features = list(
        self._client_to_data_mapping[self._client_ids[0]].keys())
    for client_id in self._client_ids:
      dataset = self._client_to_data_mapping[client_id]
      if list(dataset.keys()) != self._features:
        raise ValueError(
            f'Inconsistent features, got {list(dataset.keys())} for client {client_id}, expect {self._features}'
        )

      num_samples = dataset[self._features[0]].shape[0]
      try:
        client_datasets.assert_consistent_rows(
            self._client_dataset(client_id).all_examples())
      except ValueError as exc:
        raise ValueError(
            f'Inconsistent examples, for client {client_id}') from exc

  def slice(
      self,
      start: Optional[federated_data.ClientId] = None,
      stop: Optional[federated_data.ClientId] = None
  ) -> federated_data.FederatedData:
    if start is None and stop is None:
      client_ids = self._client_ids
    elif start is None:
      client_ids = set(i for i in self._client_ids if i < stop)
    elif stop is None:
      client_ids = set(i for i in self._client_ids if i >= start)
    else:
      client_ids = set(i for i in self._client_ids if start <= i and i < stop)
    return InMemoryFederatedData(
        {
            client_id: self._client_to_data_mapping[client_id]
            for client_id in client_ids
        }, self._preprocess_client, self._preprocess_batch)

  def num_clients(self) -> int:
    return len(self._client_ids)

  def client_ids(self) -> Iterator[federated_data.ClientId]:
    # Ids are sorted for deterministic iteration order.
    return iter(sorted(self._client_ids))

  def client_sizes(self) -> Iterator[Tuple[federated_data.ClientId, int]]:
    for client_id in self._client_ids:
      yield client_id, client_datasets.num_examples(
          self._client_dataset(client_id).all_examples(), validate=False)

  def client_size(self, client_id: federated_data.ClientId) -> int:
    return client_datasets.num_examples(
        self._client_dataset(client_id).all_examples(), validate=False)

  def clients(
      self
  ) -> Iterator[Tuple[federated_data.ClientId, client_datasets.ClientDataset]]:
    # Ids are sorted for deterministic iteration order.
    yield from self.get_clients(self._client_ids)

  def shuffled_clients(
      self,
      buffer_size: int,
      seed: Optional[int] = None
  ) -> Iterator[Tuple[federated_data.ClientId, client_datasets.ClientDataset]]:
    rng = np.random.RandomState(seed)
    while True:
      for client_id, dataset in client_datasets.buffered_shuffle(
          self.clients(), buffer_size, rng):
        yield client_id, dataset

  def get_clients(
      self, client_ids: Iterable[federated_data.ClientId]
  ) -> Iterator[Tuple[federated_data.ClientId, client_datasets.ClientDataset]]:
    for client_id in client_ids:
      yield client_id, self._client_dataset(client_id)

  def get_client(
      self,
      client_id: federated_data.ClientId) -> client_datasets.ClientDataset:
    return self._client_dataset(client_id)

  def preprocess_client(
      self, fn: Callable[[federated_data.ClientId, client_datasets.Examples],
                         client_datasets.Examples]
  ) -> federated_data.FederatedData:
    return InMemoryFederatedData(self._client_to_data_mapping,
                                 self._preprocess_client.append(fn),
                                 self._preprocess_batch)

  def preprocess_batch(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> federated_data.FederatedData:
    return InMemoryFederatedData(self._client_to_data_mapping,
                                 self._preprocess_client,
                                 self._preprocess_batch.append(fn))

  def _client_dataset(
      self,
      client_id: federated_data.ClientId) -> client_datasets.ClientDataset:
    examples = self._preprocess_client(client_id,
                                       self._client_to_data_mapping[client_id])
    return client_datasets.ClientDataset(examples, self._preprocess_batch)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.fed_avg."""
from absl.testing import absltest

from fedjax.algorithms import fed_avg
from fedjax.core import client_datasets
from fedjax.core import optimizers

import jax
import jax.numpy as jnp
import numpy.testing as npt


def grad_fn(params, batch, rng):
  del rng
  return jax.tree_util.tree_map(lambda l: l / jnp.sum(batch['x']), params)


class FedAvgTest(absltest.TestCase):

  def test_federated_averaging(self):
    client_optimizer = optimizers.sgd(learning_rate=1.0)
    server_optimizer = optimizers.sgd(learning_rate=1.0)
    client_batch_hparams = client_datasets.ShuffleRepeatBatchHParams(
        batch_size=2, num_epochs=1, seed=0)
    algorithm = fed_avg.federated_averaging(grad_fn, client_optimizer,
                                            server_optimizer,
                                            client_batch_hparams)

    with self.subTest('init'):
      state = algorithm.init({'w': jnp.array([0., 2., 4.])})
      npt.assert_array_equal(state.params['w'], [0., 2., 4.])
      self.assertLen(state.opt_state, 2)

    with self.subTest('apply'):
      clients = [
          (b'cid0',
           client_datasets.ClientDataset({'x': jnp.array([2., 4., 6.])}),
           jax.random.PRNGKey(0)),
          (b'cid1',
           client_datasets.ClientDataset({'x': jnp.array([8., 10.])}),
           jax.random.PRNGKey(1)),
      ]
      state, client_diagnostics = algorithm.apply(state, clients)
      npt.assert_allclose(state.params['w'], [0., 1.5655555, 3.131111])
      npt.assert_allclose(client_diagnostics[b'cid0']['delta_l2_norm'],
                          1.4534444262)
      npt.assert_allclose(client_diagnostics[b'cid1']['delta_l2_norm'],
                          0.2484521282)

  def test_create_train_for_each_client(self):
    client_optimizer = optimizers.sgd(learning_rate=1.0)
    train_for_each_client = fed_avg.create_train_for_each_client(
        grad_fn, client_optimizer)
    batched_clients = [
        (b'cid0',
         [{'x': jnp.array([2., 4., 6.])}, {'x': jnp.array([8., 10., 12.])}],
         jax.random.PRNGKey(0)),
        (b'cid1',
         [{'x': jnp.array([1., 3., 5.])}, {'x': jnp.array([7., 9., 11.])}],
         jax.random.PRNGKey(1)),
    ]
    server_params = {'w': jnp.array(4.0)}
    client_outputs = dict(train_for_each_client(server_params, batched_clients))
    npt.assert_allclose(client_outputs[b'cid0']['w'], 0.45555544)
    npt.assert_allclose(client_outputs[b'cid1']['w'], 0.5761316)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.client_datasets."""

import itertools

from absl.testing import absltest
from fedjax.core import client_datasets
import numpy as np
import numpy.testing as npt

# pylint: disable=g-long-lambda


class ExamplesTest(absltest.TestCase):

  def test_num_examples(self):
    with self.subTest('consistent rows'):
      self.assertEqual(
          client_datasets.num_examples({'pixels': np.zeros([10, 20, 30])}), 10)
    with self.subTest('inconsistent rows'):
      with self.assertRaises(ValueError):
        client_datasets.num_examples({
            'a': np.zeros([10, 20]),
            'b': np.zeros([20, 10])
        })

  def test_slice_examples(self):
    examples = {'a': np.arange(10), 'b': np.arange(20).reshape([10, 2])}

    with self.subTest('Slicing like [3:]'):
      sliced = client_datasets.slice_examples(examples, slice(3))
      npt.assert_equal(sliced, {'a': [0, 1, 2], 'b': [[0, 1], [2, 3], [4, 5]]})

    with self.subTest('Slicing like [3:6]'):
      sliced = client_datasets.slice_examples(examples, slice(3, 6))
      npt.assert_equal(sliced, {
          'a': [3, 4, 5],
          'b': [[6, 7], [8, 9], [10, 11]]
      })

    with self.subTest('Slicing like [3:6:2]'):
      sliced = client_datasets.slice_examples(examples, slice(3, 6, 2))
      npt.assert_equal(sliced, {'a': [3, 5], 'b': [[6, 7], [10, 11]]})

  def test_attach_mask(self):
    examples = {'a': np.arange(5), 'b': np.arange(10).reshape([5, 2])}
    mask = np.array([True, False, True, False, False])
    with self.subTest('mask key already used'):
      with self.assertRaises(ValueError):
        client_datasets.attach_mask(
            {
                **examples, client_datasets.EXAMPLE_MASK_KEY: None
            }, mask)
    with self.subTest('success'):
      npt.assert_equal(
          client_datasets.attach_mask(examples, mask), {
              'a': [0, 1, 2, 3, 4],
              'b': [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]],
              '__mask__': [True, False, True, False, False]
          })

  def test_pad_examples(self):
    examples = {'a': np.arange(5), 'b': np.arange(10).reshape([5, 2])}
    with self.subTest('maskkey already used'):
      with self.assertRaises(ValueError):
        client_datasets.pad_examples(
            {
                **examples, client_datasets.EXAMPLE_MASK_KEY: None
            }, size=8)
    with self.subTest('size too small'):
      with self.assertRaises(ValueError):
        client_datasets.pad_examples(examples, size=4)
    with self.subTest('no padding needed'):
      padded = client_datasets.pad_examples(examples, 5)
      npt.assert_equal(
          padded, {
              'a': [0, 1, 2, 3, 4],
              'b': [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]],
              '__mask__': [True, True, True, True, True]
          })
    with self.subTest('padding needed'):
      padded = client_datasets.pad_examples(examples, 6)
      npt.assert_equal(
          padded, {
              'a': [0, 1, 2, 3, 4, 0],
              'b': [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9], [0, 0]],
              '__mask__': [True, True, True, True, True, False]
          })

  def test_concat_examples(self):
    with self.subTest('empty'):
      self.assertDictEqual(client_datasets.concat_examples([]), {})

    with self.subTest('non-empty'):
      result = client_datasets.concat_examples([{
          'x': np.arange(5)
      }, {
          'x': np.arange(5, 10)
      }])
      npt.assert_equal(result, {'x': np.arange(10)})


class BatchPreprocessorTest(absltest.TestCase):

  def test_preprocessor(self):
    preprocessor = client_datasets.BatchPreprocessor([
        # Flattens `pixels`.
        lambda x: {
            **x, 'pixels': x['pixels'].reshape([-1, 28 * 28])
        },
        # Introduce `binary_label`.
        lambda x: {
            **x, 'binary_label': x['label'] % 2
        },
    ])
    fake_emnist = {
        'pixels': np.random.uniform(size=(10, 28, 28)),
        'label': np.random.randint(10, size=(10,))
    }

    with self.subTest('2 step preprocessing'):
      result = preprocessor(fake_emnist)
      npt.assert_equal(
          result, {
              'pixels': fake_emnist['pixels'].reshape([-1, 28 * 28]),
              'label': fake_emnist['label'],
              'binary_label': fake_emnist['label'] % 2
          })

  def test_append(self):
    preprocessor = client_datasets.BatchPreprocessor([
        # Flattens `pixels`.
        lambda x: {
            **x, 'pixels': x['pixels'].reshape([-1, 28 * 28])
        },
        # Introduce `binary_label`.
        lambda x: {
            **x, 'binary_label': x['label'] % 2
        },
    ])
    new_preprocessor = preprocessor.append(lambda x: {
        **x, 'sum_pixels': np.sum(x['pixels'], axis=1)
    })
    self.assertIsNot(new_preprocessor, preprocessor)
    fake_emnist = {
        'pixels': np.random.uniform(size=(10, 28, 28)),
        'label': np.random.randint(10, size=(10,))
    }
    result = new_preprocessor(fake_emnist)
    self.assertIs(result['label'], fake_emnist['label'])
    npt.assert_equal(
        result, {
            'pixels': fake_emnist['pixels'].reshape([-1, 28 * 28]),
            'label': fake_emnist['label'],
            'binary_label': fake_emnist['label'] % 2,
            'sum_pixels': np.sum(fake_emnist['pixels'], axis=(1, 2))
        })


class ClientDatasetTest(absltest.TestCase):

  def test_inconsistent_shape(self):
    with self.assertRaises(ValueError):
      client_datasets.ClientDataset({'a': np.arange(5), 'b': np.arange(4)})

  def test_len(self):
    examples = {'a': np.arange(10)}
    self.assertLen(client_datasets.ClientDataset(examples), 10)

  def test_batch(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.BatchPreprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))
    with self.subTest('keep remainder, kwargs'):
      view = d.batch(batch_size=3)
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 2)
        npt.assert_equal(batches[0], {
            'a': [0, 2, 4],
            'b': [[0, 1], [2, 3], [4, 5]]
        })
        npt.assert_equal(batches[1], {'a': [6, 8], 'b': [[6, 7], [8, 9]]})
    with self.subTest('drop remainder, hparams'):
      view = d.batch(
          client_datasets.BatchHParams(batch_size=3, drop_remainder=True))
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 1)
        npt.assert_equal(batches[0], {
            'a': [0, 2, 4],
            'b': [[0, 1], [2, 3], [4, 5]]
        })
    with self.subTest('no op drop remainder, hparams and kwargs'):
      view = d.batch(
          client_datasets.BatchHParams(batch_size=5), drop_remainder=True)
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 1)
        npt.assert_equal(batches[0], {
            'a': [0, 2, 4, 6, 8],
            'b': [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
        })

  def test_pick_final_batch_size(self):
    self.assertEqual(client_datasets._pick_final_batch_size(16, 8, 4), 8)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 1), 8)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 2), 4)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 3), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 4), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 5), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 1), 8)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 2), 4)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 3), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 4), 1)

  def test_padded_batch(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.BatchPreprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))
    with self.subTest('1 bucket, kwargs'):
      view = d.padded_batch(batch_size=3)
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 2)
        npt.assert_equal(
            batches[0], {
                'a': [0, 2, 4],
                'b': [[0, 1], [2, 3], [4, 5]],
                '__mask__': [True, True, True],
            })
        npt.assert_equal(
            batches[1], {
                'a': [6, 8, 0],
                'b': [[6, 7], [8, 9], [0, 0]],
                '__mask__': [True, True, False]
            })
    with self.subTest('2 buckets, kwargs override'):
      view = d.padded_batch(
          client_datasets.PaddedBatchHParams(batch_size=4),
          num_batch_size_buckets=2)
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 2)
        npt.assert_equal(
            batches[0], {
                'a': [0, 2, 4, 6],
                'b': [[0, 1], [2, 3], [4, 5], [6, 7]],
                '__mask__': [True, True, True, True],
            })
        npt.assert_equal(batches[1], {
            'a': [8, 0],
            'b': [[8, 9], [0, 0]],
            '__mask__': [True, False]
        })

  def test_shuffle_repeat_batch(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.BatchPreprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))
    # Number of batches under different num_epochs/num_steps combinations.
    with self.subTest('repeating'):
      self.assertLen(list(d.shuffle_repeat_batch(batch_size=5)), 1)
      self.assertLen(list(d.shuffle_repeat_batch(batch_size=3)), 2)
      self.assertLen(list(d.shuffle_repeat_batch(batch_size=1)), 5)

      self.assertEmpty(
          list(d.shuffle_repeat_batch(batch_size=7, drop_remainder=True)))
      self.assertLen(
          list(d.shuffle_repeat_batch(batch_size=5, drop_remainder=True)), 1)
      self.assertLen(
          list(d.shuffle_repeat_batch(batch_size=3, drop_remainder=True)), 1)
      self.assertLen(
          list(d.shuffle_repeat_batch(batch_size=2, drop_remainder=True)), 2)
      self.assertLen(
          list(d.shuffle_repeat_batch(batch_size=1, drop_remainder=True)), 5)

      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=5, num_epochs=None, num_steps=4)), 4)
      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=3, num_epochs=None, num_steps=4)), 4)
      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=1, num_epochs=None, num_steps=4)), 4)

      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=5,
                  num_epochs=None,
                  num_steps=4,
                  drop_remainder=True)), 4)
      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=3,
                  num_epochs=None,
                  num_steps=4,
                  drop_remainder=True)), 4)
      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=1,
                  num_epochs=None,
                  num_steps=4,
                  drop_remainder=True)), 4)

      self.assertLen(list(d.shuffle_repeat_batch(batch_size=5, num_steps=4)), 1)
      self.assertLen(list(d.shuffle_repeat_batch(batch_size=3, num_steps=4)), 2)
      self.assertLen(list(d.shuffle_repeat_batch(batch_size=1, num_steps=4)), 4)

      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=5, num_steps=4, drop_remainder=True)), 1)
      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=3, num_steps=4, drop_remainder=True)), 1)
      self.assertLen(
          list(
              d.shuffle_repeat_batch(
                  batch_size=1, num_steps=4, drop_remainder=True)), 4)

      for drop_remainder in [False, True]:
        # 100 is as good as forever.
        self.assertLen(
            list(
                itertools.islice(
                    d.shuffle_repeat_batch(
                        batch_size=3,
                        num_epochs=None,
                        drop_remainder=drop_remainder), 100)), 100)

    # Check proper shuffling.
    with self.subTest('shuffling'):
      view = d.shuffle_repeat_batch(
          batch_size=3, num_epochs=None, num_steps=4, seed=1)
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 4)
        npt.assert_equal(batches[0], {
            'a': [4, 2, 8],
            'b': [[4, 5], [2, 3], [8, 9]]
        })
        npt.assert_equal(batches[1], {
            'a': [0, 6, 4],
            'b': [[0, 1], [6, 7], [4, 5]]
        })
        npt.assert_equal(batches[2], {
            'a': [8, 6, 0],
            'b': [[8, 9], [6, 7], [0, 1]]
        })
        npt.assert_equal(batches[3], {
            'a': [2, 6, 0],
            'b': [[2, 3], [6, 7], [0, 1]]
        })

  def test_slice(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.BatchPreprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))

    with self.subTest('slice [:3]'):
      sliced = d[:3]
      batch = next(iter(sliced.batch(batch_size=3)))
      npt.assert_equal(batch, {'a': [0, 2, 4], 'b': [[0, 1], [2, 3], [4, 5]]})

    with self.subTest('slice [-3:]'):
      sliced = d[-3:]
      batch = next(iter(sliced.batch(batch_size=3)))
      npt.assert_equal(batch, {'a': [4, 6, 8], 'b': [[4, 5], [6, 7], [8, 9]]})

  def test_all_examples(self):
    raw_examples = {'a': np.arange(3), 'b': np.arange(6).reshape([3, 2])}
    with self.subTest('no preprocessing'):
      npt.assert_equal(
          client_datasets.ClientDataset(raw_examples).all_examples(),
          raw_examples)
    with self.subTest('with preprocessing'):
      npt.assert_equal(
          client_datasets.ClientDataset(
              raw_examples,
              client_datasets.BatchPreprocessor([lambda x: {
                  'c': x['a'] + 1
              }])).all_examples(), {'c': [1, 2, 3]})


class PaddedBatchClientDatasetsTest(absltest.TestCase):

  def test_empty(self):
    batches = list(
        client_datasets.padded_batch_client_datasets([], batch_size=128))
    self.assertListEqual(batches, [])

  def test_single_default_buckets(self):
    batches = list(
        client_datasets.padded_batch_client_datasets(
            [client_datasets.ClientDataset({'x': np.arange(6)})], batch_size=5))
    self.assertLen(batches, 2)
    npt.assert_equal(batches[0], {
        'x': np.arange(5),
        '__mask__': [True, True, True, True, True]
    })
    npt.assert_equal(batches[1], {
        'x': [5, 0, 0, 0, 0],
        '__mask__': [True, False, False, False, False]
    })

  def test_single_has_buckets(self):
    batches = list(
        client_datasets.padded_batch_client_datasets(
            [client_datasets.ClientDataset({'x': np.arange(8)})],
            batch_size=6,
            num_batch_size_buckets=4))
    self.assertLen(batches, 2)
    npt.assert_equal(batches[0], {
        'x': np.arange(6),
        '__mask__': [True, True, True, True, True, True]
    })
    npt.assert_equal(batches[1], {
        'x': [6, 7, 0],
        '__mask__': [True, True, False]
    })

  def test_multi(self):
    batches = list(
        client_datasets.padded_batch_client_datasets([
            client_datasets.ClientDataset({'x': np.arange(10)}),
            client_datasets.ClientDataset({'x': np.arange(10, 11)}),
            client_datasets.ClientDataset({'x': np.arange(11, 15)}),
            client_datasets.ClientDataset({'x': np.arange(15, 17)})
        ],
                                                     batch_size=4))
    self.assertLen(batches, 5)
    npt.assert_equal(batches[0], {
        'x': [0, 1, 2, 3],
        '__mask__': [True, True, True, True]
    })
    npt.assert_equal(batches[1], {
        'x': [4, 5, 6, 7],
        '__mask__': [True, True, True, True]
    })
    npt.assert_equal(batches[2], {
        'x': [8, 9, 10, 11],
        '__mask__': [True, True, True, True]
    })
    npt.assert_equal(batches[3], {
        'x': [12, 13, 14, 15],
        '__mask__': [True, True, True, True]
    })
    npt.assert_equal(batches[4], {
        'x': [16, 0, 0, 0],
        '__mask__': [True, False, False, False]
    })

  def test_preprocessor(self):
    batches = list(
        client_datasets.padded_batch_client_datasets([
            client_datasets.ClientDataset({'x': np.arange(6)},
                                          client_datasets.BatchPreprocessor(
                                              [lambda x: {
                                                  'x': x['x'] + 1
                                              }]))
        ],
                                                     batch_size=5))
    self.assertLen(batches, 2)
    npt.assert_equal(batches[0], {
        'x': np.arange(5) + 1,
        '__mask__': [True, True, True, True, True]
    })
    npt.assert_equal(batches[1], {
        'x': [6, 0, 0, 0, 0],
        '__mask__': [True, False, False, False, False]
    })

  def test_different_preprocessors(self):
    with self.assertRaisesRegex(
        ValueError,
        'client_datasets should have the identical Preprocessor object'):
      list(
          client_datasets.padded_batch_client_datasets([
              client_datasets.ClientDataset(
                  {'x': np.arange(10)}, client_datasets.BatchPreprocessor()),
              client_datasets.ClientDataset({'x': np.arange(10, 11)},
                                            client_datasets.BatchPreprocessor())
          ],
                                                       batch_size=4))

  def test_different_features(self):
    with self.assertRaisesRegex(
        ValueError, 'client_datasets should have identical features'):
      list(
          client_datasets.padded_batch_client_datasets([
              client_datasets.ClientDataset({'x': np.arange(10)}),
              client_datasets.ClientDataset({'y': np.arange(10, 11)})
          ],
                                                       batch_size=4))


class BufferedShuffleTest(absltest.TestCase):

  def test_buffered_shuffle(self):
    self.assertListEqual(
        list(
            client_datasets.buffered_shuffle(
                range(20), buffer_size=10, rng=np.random.RandomState(0))),
        [2, 1, 3, 7, 0, 13, 8, 12, 11, 17, 14, 15, 4, 9, 10, 6, 16, 18, 19, 5])


class BufferedShuffleBatchClientDatasetsTest(absltest.TestCase):

  def test_empty(self):
    batches = list(
        client_datasets.buffered_shuffle_batch_client_datasets(
            [], batch_size=5, buffer_size=10, rng=np.random.RandomState(0)))
    self.assertListEqual(batches, [])

  def test_single_buffer_1(self):
    batches = list(
        client_datasets.buffered_shuffle_batch_client_datasets(
            [client_datasets.ClientDataset({'x': np.arange(6)})],
            batch_size=5,
            buffer_size=1,
            rng=np.random.RandomState(0)))
    self.assertLen(batches, 2)
    # No shuffling.
    npt.assert_equal(batches[0], {'x': np.arange(5)})
    npt.assert_equal(batches[1], {'x': [5]})

  def test_single_buffer_4(self):
    batches = list(
        client_datasets.buffered_shuffle_batch_client_datasets(
            [client_datasets.ClientDataset({'x': np.arange(8)})],
            batch_size=6,
            buffer_size=4,
            rng=np.random.RandomState(0)))
    self.assertLen(batches, 2)
    npt.assert_equal(batches[0], {
        'x': [2, 4, 5, 6, 7, 3],
    })
    npt.assert_equal(batches[1], {
        'x': [1, 0],
    })

  def test_multi(self):
    batches = list(
        client_datasets.buffered_shuffle_batch_client_datasets(
            [
                client_datasets.ClientDataset({'x': np.arange(10)}),
                client_datasets.ClientDataset({'x': np.arange(10, 11)}),
                client_datasets.ClientDataset({'x': np.arange(11, 15)}),
                client_datasets.ClientDataset({'x': np.arange(15, 17)})
            ],
            batch_size=4,
            buffer_size=16,
            rng=np.random.RandomState(0)))
    self.assertLen(batches, 5)
    npt.assert_equal(batches[0], {
        'x': [1, 6, 16, 8],
    })
    npt.assert_equal(batches[1], {
        'x': [9, 13, 4, 2],
    })
    npt.assert_equal(batches[2], {
        'x': [14, 10, 7, 15],
    })
    npt.assert_equal(batches[3], {
        'x': [11, 3, 0, 5],
    })
    npt.assert_equal(batches[4], {
        'x': [12],
    })

  def test_preprocessor(self):
    batches = list(
        client_datasets.buffered_shuffle_batch_client_datasets(
            [
                client_datasets.ClientDataset({'x': np.arange(6)},
                                              client_datasets.BatchPreprocessor(
                                                  [lambda x: {
                                                      'x': x['x'] + 1
                                                  }]))
            ],
            batch_size=5,
            buffer_size=16,
            rng=np.random.RandomState(0)))
    self.assertLen(batches, 2)
    npt.assert_equal(batches[0], {
        'x': [6, 3, 2, 4, 1],
    })
    npt.assert_equal(batches[1], {
        'x': [5],
    })

  def test_different_preprocessors(self):
    with self.assertRaisesRegex(
        ValueError,
        'client_datasets should have the identical Preprocessor object'):
      list(
          client_datasets.buffered_shuffle_batch_client_datasets(
              [
                  client_datasets.ClientDataset(
                      {'x': np.arange(10, 20)},
                      client_datasets.BatchPreprocessor()),
                  client_datasets.ClientDataset(
                      {'x': np.arange(20, 30)},
                      client_datasets.BatchPreprocessor())
              ],
              batch_size=4,
              buffer_size=16,
              rng=np.random.RandomState(0)))

  def test_different_features(self):
    with self.assertRaisesRegex(
        ValueError, 'client_datasets should have identical features'):
      list(
          client_datasets.buffered_shuffle_batch_client_datasets(
              [
                  client_datasets.ClientDataset({'x': np.arange(10)}),
                  client_datasets.ClientDataset({'y': np.arange(10, 11)})
              ],
              batch_size=4,
              buffer_size=16,
              rng=np.random.RandomState(0)))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.datasets.emnist.

This file only tests preprocessing functions.
"""

from absl.testing import absltest
from fedjax.datasets import emnist
import numpy as np
import numpy.testing as npt


class EmnistTest(absltest.TestCase):

  def test_domain_id(self):
    with self.subTest('invalid'):
      with self.assertRaisesRegex(ValueError, 'Invalid client_id'):
        emnist.domain_id(b'not_emnist')

    with self.subTest('domain 0'):
      for client_id in [b'f2100_xx', b'f2222_yy', b'f2599_zz']:
        self.assertEqual(emnist.domain_id(client_id), 0)
        self.assertEqual(emnist.domain_id(b'A' * 16 + b':' + client_id), 0)

    with self.subTest('domain 1'):
      for client_id in [b'f0000_xx', b'f2099_yy', b'f2600_zz', b'f9999_99']:
        self.assertEqual(emnist.domain_id(client_id), 1)
        self.assertEqual(emnist.domain_id(b'A' * 16 + b':' + client_id), 1)

  def test_preprocess(self):
    pixels = np.random.uniform(size=[128, 28, 28])
    label = np.random.randint(10, size=[128])
    npt.assert_equal(
        emnist.preprocess_client(b'f0000_xx', {
            'pixels': pixels,
            'label': label
        }), {
            'pixels': pixels,
            'label': label,
            'domain_id': np.ones_like(label)
        })

    domain_id = np.random.randint(2, size=[128])
    npt.assert_equal(
        emnist.preprocess_batch({
            'pixels': pixels,
            'label': label,
            'domain_id': domain_id
        }), {
            'x': 1 - pixels[..., np.newaxis],
            'y': label,
            'domain_id': domain_id
        })


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax."""

import unittest

import fedjax


class FedjaxTest(unittest.TestCase):
  """Test fedjax can be imported correctly."""

  def test_import(self):
    self.assertTrue(hasattr(fedjax, 'FederatedAlgorithm'))
    self.assertTrue(hasattr(fedjax.aggregators, 'Aggregator'))
    self.assertTrue(hasattr(fedjax.algorithms, 'fed_avg'))
    self.assertTrue(hasattr(fedjax.datasets, 'emnist'))
    self.assertTrue(hasattr(fedjax.models, 'emnist'))
    self.assertTrue(hasattr(fedjax.training, 'save_checkpoint'))

  def test_no_core(self):
    self.assertFalse(hasattr(fedjax, 'core'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mime implementation.

Based on the paper:

Mime: Mimicking Centralized Stochastic Algorithms in Federated Learning
    Sai Praneeth Karimireddy, Martin Jaggi, Satyen Kale, Mehryar Mohri,
    Sashank J. Reddi, Sebastian U. Stich, Ananda Theertha Suresh
    https://arxiv.org/abs/2008.03606
"""

from typing import List

from fedjax.legacy import core
from fedjax.legacy.algorithms import mime_lite


class Mime(core.FederatedAlgorithm):
  """Mime algorithm."""

  def __init__(self,
               federated_data: core.FederatedData,
               model: core.Model,
               base_optimizer: core.Optimizer,
               hparams: mime_lite.MimeLiteHParams,
               rng_seq: core.PRNGSequence):
    """Initializes MimeLite algorithm.

    Args:
      federated_data: Federated data separated per client.
      model: Model implementation.
      base_optimizer: Base centralized optimizer which we will try mimic.
      hparams: Hyperparameters for federated averaging.
      rng_seq: Iterator of JAX random keys.
    """
    self._federated_data = federated_data
    self._model = model
    self._base_optimizer = base_optimizer
    self._hparams = hparams
    self._rng_seq = rng_seq
    self._client_trainer = core.ControlVariateTrainer(model, base_optimizer)

  @property
  def federated_data(self) -> core.FederatedData:
    return self._federated_data

  @property
  def model(self) -> core.Model:
    return self._model

  def init_state(self) -> mime_lite.MimeLiteState:
    params = self._model.init_params(next(self._rng_seq))
    opt_state = self._base_optimizer.init_fn(params)
    return mime_lite.MimeLiteState(params, opt_state)

  def run_round(self, state: mime_lite.MimeLiteState,
                client_ids: List[str]) -> mime_lite.MimeLiteState:
    """Runs one round of Mime."""
    # Compute full-batch gradient at server params on train data.
    combined_dataset = core.preprocess_tf_dataset(
        core.create_tf_dataset_for_clients(self.federated_data, client_ids),
        self._hparams.combined_data_hparams)
    server_grads = mime_lite.compute_gradient(
        stream=combined_dataset.as_numpy_iterator(),
        params=state.params,
        model=self._model,
        rng_seq=self._rng_seq,
    )

    # Train on clients using custom ControlVariateTrainer.
    client_states = core.train_multiple_clients(
        federated_data=self.federated_data,
        client_ids=client_ids,
        client_trainer=self._client_trainer,
        init_client_trainer_state=self._client_trainer.init_state(
            params=state.params,
            opt_state=state.opt_state,
            control_variate=server_grads),
        rng_seq=self._rng_seq,
        client_data_hparams=self._hparams.train_data_hparams)

    # Weighted average of param delta across clients.
    def select_delta_params_and_weight(client_state):
      delta_params = core.tree_multimap(lambda a, b: a - b, state.params,
                                        client_state.params)
      return delta_params, client_state.num_examples

    delta_params_and_weight = map(select_delta_params_and_weight, client_states)
    delta_params = core.tree_mean(delta_params_and_weight)

    # Server params uses weighted average of client updates, scaled by the
    # server_learning_rate.
    params = core.tree_multimap(
        lambda p, q: p - self._hparams.server_learning_rate * q, state.params,
        delta_params)
    # Update server opt_state using base_optimimzer and server gradient.
    _, opt_state = self._base_optimizer.update_fn(server_grads, state.opt_state)
    return mime_lite.MimeLiteState(params, opt_state)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.training.federated_experiment."""

import os.path

from absl.testing import absltest
from fedjax.core import client_datasets
from fedjax.core import client_samplers
from fedjax.core import federated_algorithm
from fedjax.core import in_memory_federated_data
from fedjax.core import metrics
from fedjax.core import models
from fedjax.training import federated_experiment
import jax
import numpy as np
import numpy.testing as npt
import tensorflow as tf


class FakeClientSampler(client_samplers.ClientSampler):
  """Sequentially creates single client samples."""

  def __init__(self, base=0):
    self._base = base
    self._round_num = 0

  def set_round_num(self, round_num):
    self._round_num = round_num

  def sample(self):
    client_id = self._round_num + self._base
    dataset = client_datasets.ClientDataset({
        'x': np.array([client_id], dtype=np.int32),
        'y': np.array([client_id], dtype=np.int32) % 2
    })
    rng = None
    self._round_num += 1
    return [(client_id, dataset, rng)]


def fake_algorithm():
  """Counts the number of clients and sums up the 'x' feature."""

  def init():
    # num_clients, sum_values
    return 0, 0

  def apply(state, clients):
    num_clients, sum_values = state
    for _, dataset, _ in clients:
      num_clients += 1
      for x in dataset.all_examples()['x']:
        sum_values += x
    state = num_clients, sum_values
    return state, None

  return federated_algorithm.FederatedAlgorithm(init, apply)


class FakeEvaluationFn(federated_experiment.EvaluationFn):

  def __init__(self, test, expected_state):
    self._test = test
    self._expected_state = expected_state

  def __call__(self, state, round_num):
    self._test.assertEqual(state, self._expected_state[round_num])
    return {'round_num': round_num}


class FakeTrainClientsEvaluationFn(federated_experiment.TrainClientsEvaluationFn
                                  ):

  def __init__(self, test, expected_state, expected_client_ids):
    self._test = test
    self._expected_state = expected_state
    self._expected_client_ids = expected_client_ids

  def __call__(self, state, round_num, train_clients):
    self._test.assertEqual(state, self._expected_state[round_num])
    client_ids = [i[0] for i in train_clients]
    self._test.assertCountEqual(client_ids,
                                self._expected_client_ids[round_num])
    return {'round_num': round_num}


class RunFederatedExperimentTest(absltest.TestCase):

  def test_no_eval(self):
    config = federated_experiment.FederatedExperimentConfig(
        root_dir=self.create_tempdir(), num_rounds=5)
    client_sampler = FakeClientSampler()
    algorithm = fake_algorithm()
    state = federated_experiment.run_federated_experiment(
        algorithm=algorithm,
        init_state=algorithm.init(),
        client_sampler=client_sampler,
        config=config)
    self.assertEqual(state, (5, 15))

  def test_checkpoint(self):
    with self.subTest('checkpoint init'):
      config = federated_experiment.FederatedExperimentConfig(
          root_dir=self.create_tempdir(), num_rounds=5, checkpoint_frequency=3)
      client_sampler = FakeClientSampler()
      algorithm = fake_algorithm()
      state = federated_experiment.run_federated_experiment(
          algorithm=algorithm,
          init_state=(1, -1),
          client_sampler=client_sampler,
          config=config)
      self.assertEqual(state, (6, 14))
      self.assertCountEqual(
          tf.io.gfile.glob(os.path.join(config.root_dir, 'checkpoint_*')),
          [os.path.join(config.root_dir, 'checkpoint_00000003')])

    with self.subTest('checkpoint restore'):
      # Restored state is (4, 5). FakeSampler produces clients [5, 6].
      state = federated_experiment.run_federated_experiment(
          algorithm=algorithm,
          init_state=None,
          client_sampler=FakeClientSampler(1),
          config=config)
      self.assertEqual(state, (6, 16))
      self.assertCountEqual(
          tf.io.gfile.glob(os.path.join(config.root_dir, 'checkpoint_*')),
          [os.path.join(config.root_dir, 'checkpoint_00000004')])

  def test_periodic_eval_fn_map(self):
    config = federated_experiment.FederatedExperimentConfig(
        root_dir=self.create_tempdir(), num_rounds=5, eval_frequency=3)
    client_sampler = FakeClientSampler()
    algorithm = fake_algorithm()
    expected_state = {1: (1, 1), 3: (3, 6)}
    expected_client_ids = {1: [1], 3: [3]}
    state = federated_experiment.run_federated_experiment(
        algorithm=algorithm,
        init_state=algorithm.init(),
        client_sampler=client_sampler,
        config=config,
        periodic_eval_fn_map={
            'test_eval':
                FakeEvaluationFn(self, expected_state),
            'train_eval':
                FakeTrainClientsEvaluationFn(self, expected_state,
                                             expected_client_ids)
        })
    self.assertEqual(state, (5, 15))
    self.assertCountEqual(
        tf.io.gfile.glob(os.path.join(config.root_dir, '*eval*')), [
            os.path.join(config.root_dir, 'test_eval'),
            os.path.join(config.root_dir, 'train_eval')
        ])

  def test_final_eval_fn_map(self):
    config = federated_experiment.FederatedExperimentConfig(
        root_dir=self.create_tempdir(), num_rounds=5, eval_frequency=3)
    client_sampler = FakeClientSampler()
    algorithm = fake_algorithm()
    expected_state = {5: (5, 15)}
    state = federated_experiment.run_federated_experiment(
        algorithm=algorithm,
        init_state=algorithm.init(),
        client_sampler=client_sampler,
        config=config,
        final_eval_fn_map={
            'final_eval': FakeEvaluationFn(self, expected_state)
        })
    self.assertEqual(state, (5, 15))
    self.assertCountEqual(
        tf.io.gfile.glob(os.path.join(config.root_dir, 'final_eval.tsv')),
        [os.path.join(config.root_dir, 'final_eval.tsv')])


def fake_model():

  def apply_for_eval(params, example):
    del params
    return jax.nn.one_hot(example['x'] % 3, 3)

  eval_metrics = {'accuracy': metrics.Accuracy()}
  return models.Model(
      init=None,
      apply_for_train=None,
      apply_for_eval=apply_for_eval,
      train_loss=None,
      eval_metrics=eval_metrics)


class FakeState:
  params = None


class EvaluationFnsTest(absltest.TestCase):

  def test_model_sample_clients_evaluation_fn(self):
    eval_fn = federated_experiment.ModelSampleClientsEvaluationFn(
        FakeClientSampler(), fake_model(),
        client_datasets.PaddedBatchHParams(batch_size=4))
    state = FakeState()
    npt.assert_equal(eval_fn(state, 1), {'accuracy': np.array(1.)})
    npt.assert_equal(eval_fn(state, 2), {'accuracy': np.array(0.)})
    npt.assert_equal(eval_fn(state, 3), {'accuracy': np.array(0.)})
    npt.assert_equal(eval_fn(state, 4), {'accuracy': np.array(0.)})
    npt.assert_equal(eval_fn(state, 5), {'accuracy': np.array(0.)})
    npt.assert_equal(eval_fn(state, 6), {'accuracy': np.array(1.)})

  def test_model_full_evaluation_fn(self):
    sampler = FakeClientSampler()
    sampler.set_round_num(1)
    clients = [sampler.sample()[0] for _ in range(4)]
    fd = in_memory_federated_data.InMemoryFederatedData(
        dict((k, v.all_examples()) for k, v, _ in clients))
    eval_fn = federated_experiment.ModelFullEvaluationFn(
        fd, fake_model(), client_datasets.PaddedBatchHParams(batch_size=4))
    state = FakeState()
    npt.assert_equal(eval_fn(state, 1), {'accuracy': np.array(0.25)})
    npt.assert_equal(eval_fn(state, 100), {'accuracy': np.array(0.25)})

  def test_model_train_clients_evaluation_fn(self):
    sampler = FakeClientSampler()
    sampler.set_round_num(1)
    clients = [sampler.sample()[0] for _ in range(4)]
    eval_fn = federated_experiment.ModelTrainClientsEvaluationFn(
        fake_model(), client_datasets.PaddedBatchHParams(batch_size=4))
    state = FakeState()
    npt.assert_equal(eval_fn(state, 1, clients), {'accuracy': np.array(0.25)})
    npt.assert_equal(eval_fn(state, 100, clients), {'accuracy': np.array(0.25)})


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated algorithm interface."""

from typing import Any, Callable, Mapping, Sequence, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_data
from fedjax.core.typing import PRNGKey
from fedjax.core.typing import PyTree


# Server state contains any round specific parameters (e.g. model parameters)
# that will be passed from round to round. This state must be serializable for
# checkpointing. We strongly recommend making this a PyTree.
ServerState = PyTree


@dataclasses.dataclass
class FederatedAlgorithm:
  """Container for all federated algorithms.

  FederatedAlgorithm defines the required methods that need to be implemented
  by all custom federated algorithms. Defining federated algorithms in this
  structure will allow implementations to work seamlessly with the convenience
  methods in the ``fedjax.training`` API, like checkpointing.

  Example toy implementation::

    # Federated algorithm that just counts total number of points across clients
    # across rounds.

    def count_federated_algorithm():

      def init(init_count):
        return {'count': init_count}

      def apply(state, clients):
        count = 0
        client_diagnostics = {}
        # Count sizes across clients.
        for client_id, client_dataset, _ in clients:
          # Summation across clients in one round.
          client_count = len(client_dataset)
          count += client_count
          client_diagnostics[client_id] = client_count
        # Summation across rounds.
        state = {'count': state['count'] + count}
        return state, client_diagnostics

      return FederatedAlgorithm(init, apply)

    rng = jax.random.PRNGKey(0)  # Unused.
    all_clients = [
        [
          (b'cid0', ClientDataset({'x': jnp.array([1, 2, 1, 2, 3, 4])}), rng),
          (b'cid1', ClientDataset({'x': jnp.array([1, 2, 3, 4, 5])}), rng),
          (b'cid2', ClientDataset({'x': jnp.array([1, 1, 2])}), rng),
        ],
        [
          (b'cid3', ClientDataset({'x': jnp.array([1, 2, 3, 4])}), rng),
          (b'cid4', ClientDataset({'x': jnp.array([1, 1, 2, 1, 2, 3])}), rng),
          (b'cid5', ClientDataset({'x': jnp.array([1, 2, 3, 4, 5, 6, 7])}), rng),
        ],
    ]
    algorithm = count_federated_algorithm()
    state = algorithm.init(0)
    for round_num in range(2):
      state, client_diagnostics = algorithm.apply(state, all_clients[round_num])
      print(round_num, state)
      print(round_num, client_diagnostics)
    # 0 {'count': 14}
    # 0 {b'cid0': 6, b'cid1': 5, b'cid2': 3}
    # 1 {'count': 31}
    # 1 {b'cid3': 4, b'cid4': 6, b'cid5': 7}

  Attributes:
    init: Initializes the ``ServerState``. Typically, the input to this method
      will be the initial model ``Params``. This should only be run once at the
      beginning of training.
    apply: Completes one round of federated training given an input
      ``ServerState`` and a sequence of tuples of client identifier, client
      dataset, and client rng. The output will be a new, updated ``ServerState``
      and accumulated per step results keyed by client identifier
      (e.g. train metrics).
  """
  init: Callable[..., ServerState]
  apply: Callable[[
      ServerState, Sequence[Tuple[federated_data.ClientId,
                                  client_datasets.ClientDataset, PRNGKey]]
  ], Tuple[ServerState, Mapping[federated_data.ClientId, Any]]]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.tree_util."""

from absl.testing import absltest

from fedjax.core import tree_util

import jax.numpy as jnp
import numpy.testing as npt


class TreeUtilTest(absltest.TestCase):

  def test_tree_weight(self):
    pytree = {
        'x': jnp.array([[[4, 5]], [[1, 1]]]),
        'y': jnp.array([[3], [1]]),
    }
    weight = 2.0
    weight_pytree = tree_util.tree_weight(pytree, weight)
    npt.assert_array_equal(weight_pytree['x'], [[[8.0, 10.0]], [[2.0, 2.0]]])
    npt.assert_array_equal(weight_pytree['y'], [[6.0], [2.0]])

  def test_tree_sum(self):
    pytree_1 = {
        'x': jnp.array([[[4, 5]], [[1, 1]]]),
        'y': jnp.array([[3], [1]]),
    }
    pytree_2 = {
        'x': jnp.array([[[2, 3]], [[4, 5]]]),
        'y': jnp.array([[6], [7]]),
    }
    pytree = tree_util.tree_sum([pytree_1, pytree_2])
    npt.assert_array_equal(pytree['x'], [[[6, 8]], [[5, 6]]])
    npt.assert_array_equal(pytree['y'], [[9], [8]])

  def test_tree_mean(self):
    pytrees = [(0, 1), (2, 3), (4, 5)]
    weights = [6., 7., 8.]
    pytrees_and_weights = zip(pytrees, weights)
    pytree = tree_util.tree_mean(pytrees_and_weights)
    npt.assert_array_almost_equal(pytree,
                                  (2.1904761904761907, 3.1904761904761907))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Derived from https://github.com/google/flax/blob/master/flax/serialization.py
#
# Copyright 2021 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple serialization scheme based on msgpack.

We support serializing a dict and list of ndarrays holding numeric types, and
bytes (note tuples cannot be serialized since msgpack does not distinguish
tuples from lists during deserialization).
"""

import enum
import pickle

from absl import logging
import jax
import msgpack
import numpy as np
# TODO(b/188556866): Remove dependency on TensorFlow.
import tensorflow as tf


def save_state(state, path):
  """Saves state to file path."""
  logging.info('Saving state to %s.', path)
  with tf.io.gfile.GFile(path, 'wb') as f:
    pickle.dump(state, f)


def load_state(path):
  """Loads saved state from file path."""
  logging.info('Loading params from %s.', path)
  with tf.io.gfile.GFile(path, 'rb') as f:
    return pickle.load(f)


# On-the-wire / disk serialization format

# We encode state-dicts via msgpack, using its custom type extension.
# https://github.com/msgpack/msgpack/blob/master/spec.md
#
# - ndarrays and DeviceArrays of numeric types are serialized to nested
#   msgpack-encoded string of (shape-tuple, dtype-name (e.g. 'float32'),
#   row-major array-bytes).
#
# - native complex scalars are converted to nested msgpack-encoded tuples
#   (real, imag).
#
# - ndarrays of bytes are serialized to nested msgpack-encoded string of
#   (shape-tuple, concatenated bytes).


def _ndarray_to_bytes(arr):
  """Save ndarray to simple msgpack encoding."""
  if isinstance(arr, jax.xla.DeviceArray):
    arr = np.array(arr)
  if arr.dtype.hasobject or arr.dtype.isalignedstruct:
    raise ValueError('Object and structured dtypes not supported '
                     'for serialization of ndarrays.')
  tpl = (arr.shape, arr.dtype.name, arr.tobytes('C'))
  return msgpack.packb(tpl, use_bin_type=True)


def _dtype_from_name(name):
  """Handle JAX bfloat16 dtype correctly."""
  if name == b'bfloat16':
    return jax.numpy.bfloat16
  else:
    return np.dtype(name)


def _ndarray_from_bytes(data):
  """Load ndarray from simple msgpack encoding."""
  shape, dtype_name, buffer = msgpack.unpackb(data, raw=True)
  return np.frombuffer(buffer,
                       dtype=_dtype_from_name(dtype_name),
                       count=-1,
                       offset=0).reshape(shape, order='C')


def _bytes_ndarray_to_bytes(x):
  shape = x.shape
  flat = list(x.flatten())
  if flat and not isinstance(flat[0], bytes):
    raise ValueError('Only ndarrays holding bytes objects can be serialized.')
  tpl = shape, flat
  return msgpack.packb(tpl, use_bin_type=True)


def _object_ndarray_from_bytes(data):
  shape, flat = msgpack.unpackb(data, raw=True)
  return np.array(flat, dtype=object).reshape(shape)


class _MsgpackExtType(enum.IntEnum):
  """Messagepack custom type ids."""
  ndarray = 1
  native_complex = 2
  npscalar = 3
  bytes_ndarray = 4


def _msgpack_ext_pack(x):
  """Messagepack encoders for custom types."""
  if isinstance(x, np.ndarray) and x.dtype.hasobject:
    return msgpack.ExtType(_MsgpackExtType.bytes_ndarray,
                           _bytes_ndarray_to_bytes(x))
  elif isinstance(x, (np.ndarray, jax.xla.DeviceArray)):
    return msgpack.ExtType(_MsgpackExtType.ndarray, _ndarray_to_bytes(x))
  elif np.issctype(type(x)):
    # pack scalar as ndarray
    return msgpack.ExtType(_MsgpackExtType.npscalar,
                           _ndarray_to_bytes(np.asarray(x)))
  elif isinstance(x, complex):
    return msgpack.ExtType(_MsgpackExtType.native_complex,
                           msgpack.packb((x.real, x.imag)))
  print('falling back', repr(x))
  return x


def _msgpack_ext_unpack(code, data):
  """Messagepack decoders for custom types."""
  if code == _MsgpackExtType.ndarray:
    return _ndarray_from_bytes(data)
  elif code == _MsgpackExtType.native_complex:
    complex_tuple = msgpack.unpackb(data)
    return complex(complex_tuple[0], complex_tuple[1])
  elif code == _MsgpackExtType.npscalar:
    ar = _ndarray_from_bytes(data)
    return ar[()]  # unpack ndarray to scalar
  elif code == _MsgpackExtType.bytes_ndarray:
    return _object_ndarray_from_bytes(data)
  return msgpack.ExtType(code, data)


# User-facing API calls:


def msgpack_serialize(pytree):
  """Save data structure to bytes in msgpack format.

  Low-level function that only supports python trees with array leaves,
  for custom objects use `to_bytes`.

  Args:
    pytree: python tree of dict, list with python primitives and array leaves.

  Returns:
    msgpack-encoded bytes of pytree.
  """
  return msgpack.packb(pytree, default=_msgpack_ext_pack, strict_types=True)


def msgpack_deserialize(encoded_pytree):
  """Restore data structure from bytes in msgpack format.

  Low-level function that only supports python trees with array leaves,
  for custom objects use `from_bytes`.

  Args:
    encoded_pytree: msgpack-encoded bytes of python tree.

  Returns:
    Python tree of dict, list with python primitive and array leaves.
  """
  return msgpack.unpackb(
      encoded_pytree, ext_hook=_msgpack_ext_unpack, raw=False)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.mime."""

from absl.testing import absltest

from fedjax.algorithms import mime
from fedjax.core import client_datasets
from fedjax.core import models
from fedjax.core import optimizers

import jax
import jax.numpy as jnp
import numpy.testing as npt


def per_example_loss(params, batch, rng):
  del rng
  return batch['x'] * params['w']


grad_fn = models.grad(per_example_loss)


class MimeTest(absltest.TestCase):

  def test_mime(self):
    base_optimizer = optimizers.sgd(learning_rate=1.0)
    train_batch_hparams = client_datasets.ShuffleRepeatBatchHParams(
        batch_size=2, num_epochs=1, seed=0)
    grad_batch_hparams = client_datasets.PaddedBatchHParams(batch_size=2)
    server_learning_rate = 0.2
    algorithm = mime.mime(per_example_loss, base_optimizer, train_batch_hparams,
                          grad_batch_hparams, server_learning_rate)

    with self.subTest('init'):
      state = algorithm.init({'w': jnp.array(4.)})
      npt.assert_equal(state.params, {'w': jnp.array(4.)})
      self.assertLen(state.opt_state, 2)

    with self.subTest('apply'):
      clients = [
          (b'cid0',
           client_datasets.ClientDataset({'x': jnp.array([2., 4., 6.])}),
           jax.random.PRNGKey(0)),
          (b'cid1', client_datasets.ClientDataset({'x': jnp.array([8., 10.])}),
           jax.random.PRNGKey(1)),
      ]
      state, client_diagnostics = algorithm.apply(state, clients)
      npt.assert_allclose(state.params['w'], 2.08)
      npt.assert_allclose(client_diagnostics[b'cid0']['delta_l2_norm'], 12.)
      npt.assert_allclose(client_diagnostics[b'cid1']['delta_l2_norm'], 6.)

  def test_create_grads_for_each_client(self):
    grads_for_each_client = mime.create_grads_for_each_client(grad_fn)
    server_params = {'w': jnp.array(4.)}
    padded_batch_clients = [
        (b'cid0', [{
            'x': jnp.array([2., 4.]),
            '__mask__': jnp.array([True, True])
        }, {
            'x': jnp.array([6., 0.]),
            '__mask__': jnp.array([True, False])
        }], jax.random.PRNGKey(0)),
        (b'cid1', [{
            'x': jnp.array([8., 10.]),
            '__mask__': jnp.array([True, True])
        }], jax.random.PRNGKey(1)),
    ]
    npt.assert_equal(
        list(grads_for_each_client(server_params, padded_batch_clients)),
        [(b'cid0', ({'w': jnp.array(12.)}, jnp.array(3.))),
         (b'cid1', ({'w': jnp.array(18.)}, jnp.array(2.)))])

  def test_create_train_for_each_client(self):
    base_optimizer = optimizers.sgd(learning_rate=1.0)
    train_for_each_client = mime.create_train_for_each_client(
        grad_fn, base_optimizer)
    batch_clients = [
        (b'cid0', [{
            'x': jnp.array([6., 4.])
        }, {
            'x': jnp.array([2., 2.])
        }], jax.random.PRNGKey(0)),
        (b'cid1', [{
            'x': jnp.array([10., 8])
        }], jax.random.PRNGKey(1)),
    ]
    server_params = {'w': jnp.array(4.)}
    server_opt_state = base_optimizer.init(server_params)
    server_grads = {'w': jnp.array(6.)}
    shared_input = {
        'params': server_params,
        'opt_state': server_opt_state,
        'control_variate': server_grads
    }
    client_outputs = dict(train_for_each_client(shared_input, batch_clients))
    npt.assert_allclose(client_outputs[b'cid0']['w'], 12.)
    npt.assert_allclose(client_outputs[b'cid1']['w'], 6.)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library for defining regularizers."""

from typing import Callable, Optional

from fedjax.core.typing import Params

import jax
import jax.numpy as jnp


def _l2_regularize(params: Params, weight: float,
                   center_params: Optional[Params],
                   params_weights: Optional[Params]) -> float:
  """Returns L2 regularization weight."""
  if center_params is not None:
    params = jax.tree_multimap(lambda a, b: a - b, params, center_params)
  leaves = jax.tree_util.tree_leaves(params)
  if params_weights is not None:
    pw_leaves = jax.tree_util.tree_leaves(params_weights)
    return sum(jnp.vdot(pw, jnp.square(x))
               for pw, x in zip(pw_leaves, leaves)) * weight
  return sum(jnp.vdot(x, x) for x in leaves) * weight


def l2_regularizer(
    weight: float,
    center_params: Optional[Params] = None,
    params_weights: Optional[Params] = None) -> Callable[[Params], float]:
  """Returns L2 regularization function.

  Args:
    weight: Weight applied to L2 norm.
    center_params: Model parameter values to regularize toward.
    params_weights: Per-parameter weighting, which allows different
      regularization strengths for each parameter.

  Returns:
    L2 regularization function typically used in calculating training loss.
  """

  def func(params):
    return _l2_regularize(params, weight, center_params, params_weights)

  return func



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.datasets.downloads.

Can only be run locally because of network access.
"""

import binascii
import hashlib
import os
import os.path
import shutil

from absl import flags
from absl.testing import absltest
from fedjax.datasets import downloads

FLAGS = flags.FLAGS


class DownloadsTest(absltest.TestCase):
  SOURCE = 'https://storage.googleapis.com/tff-datasets-public/shakespeare.sqlite.lzma'
  HEXDIGEST = b'd3d11fceb9e105439ac6f4d52af6efafed5a2a1e1eb24c5bd2dd54ced242f5c4'

  def validate_file(self, path):
    h = hashlib.sha256()
    with open(path, 'rb') as f:
      data = f.read()
      h.update(data)
    self.assertEqual(binascii.hexlify(h.digest()), self.HEXDIGEST)
    # Don't use assertLen. It dumps the entire bytes object when failing.
    length = len(data)
    self.assertEqual(length, 1329828)

  def setUp(self):
    super().setUp()
    # This will be set in actual tests.
    self._cache_dir = None

  def tearDown(self):
    super().tearDown()
    if self._cache_dir is not None:
      shutil.rmtree(self._cache_dir)

  def _with_cache_dir(self, cache_dir, expected_path):
    # First access, being copied.
    path = downloads.maybe_download(self.SOURCE, cache_dir)
    self._cache_dir = os.path.dirname(path)
    self.assertEqual(path, expected_path)
    self.validate_file(path)
    stat = os.stat(path)
    # ctime >= mtime because of renaming.
    self.assertGreaterEqual(stat.st_ctime_ns, stat.st_mtime_ns)

    # Second access, no copy.
    path1 = downloads.maybe_download(self.SOURCE, cache_dir)
    self.assertEqual(path1, path)
    self.validate_file(path1)
    # Check that the file hasn't been overwritten.
    stat1 = os.stat(path1)
    self.assertEqual(stat1.st_mtime_ns, stat.st_mtime_ns)

  def test_default(self):
    self._with_cache_dir(
        cache_dir=None,
        expected_path=os.path.join(
            os.path.expanduser('~'), '.cache/fedjax/shakespeare.sqlite.lzma'))

  def test_cache_dir(self):
    cache_dir = os.path.join(FLAGS.test_tmpdir, 'test_cache_dir')
    cache_file = os.path.join(cache_dir, 'shakespeare.sqlite.lzma')
    self._with_cache_dir(cache_dir=cache_dir, expected_path=cache_file)

  def test_partial_file(self):
    cache_dir = os.path.join(FLAGS.test_tmpdir, 'test_partial_file')
    cache_file = os.path.join(cache_dir, 'shakespeare.sqlite.lzma')
    os.makedirs(cache_dir, exist_ok=True)
    with open(cache_file + '.partial', 'w') as f:
      f.write('hel')
    self._with_cache_dir(cache_dir=cache_dir, expected_path=cache_file)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated averaging implementation using fedjax.core.

This is the more performant implementation that matches what would be used in
the :py:mod:`fedjax.algorithms.fed_avg` . The key difference between this and
the basic version is the use of :py:mod:`fedjax.core.for_each_client`

Communication-Efficient Learning of Deep Networks from Decentralized Data
    H. Brendan McMahan, Eider Moore, Daniel Ramage,
    Seth Hampson, Blaise Aguera y Arcas. AISTATS 2017.
    https://arxiv.org/abs/1602.05629
"""

from typing import Any, Callable, Mapping, Sequence, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import optimizers
from fedjax.core import tree_util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey

import jax

Grads = Params


def create_train_for_each_client(grad_fn, client_optimizer):
  """Builds client_init, client_step, client_final for for_each_client."""

  def client_init(server_params, client_rng):
    opt_state = client_optimizer.init(server_params)
    client_step_state = {
        'params': server_params,
        'opt_state': opt_state,
        'rng': client_rng,
    }
    return client_step_state

  def client_step(client_step_state, batch):
    rng, use_rng = jax.random.split(client_step_state['rng'])
    grads = grad_fn(client_step_state['params'], batch, use_rng)
    opt_state, params = client_optimizer.apply(grads,
                                               client_step_state['opt_state'],
                                               client_step_state['params'])
    next_client_step_state = {
        'params': params,
        'opt_state': opt_state,
        'rng': rng,
    }
    return next_client_step_state

  def client_final(server_params, client_step_state):
    delta_params = jax.tree_util.tree_multimap(lambda a, b: a - b,
                                               server_params,
                                               client_step_state['params'])
    return delta_params

  return for_each_client.for_each_client(client_init, client_step, client_final)


@dataclasses.dataclass
class ServerState:
  """State of server passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    opt_state: A pytree representing the server optimizer state.
  """
  params: Params
  opt_state: optimizers.OptState


def federated_averaging(
    grad_fn: Callable[[Params, BatchExample, PRNGKey], Grads],
    client_optimizer: optimizers.Optimizer,
    server_optimizer: optimizers.Optimizer,
    client_batch_hparams: client_datasets.ShuffleRepeatBatchHParams
) -> federated_algorithm.FederatedAlgorithm:
  """Builds federated averaging.

  Args:
    grad_fn: A function from (params, batch_example, rng) to gradients.
      This can be created with :func:`fedjax.core.model.model_grad`.
    client_optimizer: Optimizer for local client training.
    server_optimizer: Optimizer for server update.
    client_batch_hparams: Hyperparameters for batching client dataset for train.

  Returns:
    FederatedAlgorithm
  """
  train_for_each_client = create_train_for_each_client(grad_fn,
                                                       client_optimizer)

  def init(params: Params) -> ServerState:
    opt_state = server_optimizer.init(params)
    return ServerState(params, opt_state)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]]
  ) -> Tuple[ServerState, Mapping[federated_data.ClientId, Any]]:
    client_num_examples = {cid: len(cds) for cid, cds, _ in clients}
    batch_clients = [(cid, cds.shuffle_repeat_batch(client_batch_hparams), crng)
                     for cid, cds, crng in clients]
    client_diagnostics = {}
    # Running weighted mean of client updates. We do this iteratively to avoid
    # loading all the client outputs into memory since they can be prohibitively
    # large depending on the model parameters size.
    delta_params_sum = tree_util.tree_zeros_like(server_state.params)
    num_examples_sum = 0.
    for client_id, delta_params in train_for_each_client(
        server_state.params, batch_clients):
      num_examples = client_num_examples[client_id]
      delta_params_sum = tree_util.tree_add(
          delta_params_sum, tree_util.tree_weight(delta_params, num_examples))
      num_examples_sum += num_examples
      # We record the l2 norm of client updates as an example, but it is not
      # required for the algorithm.
      client_diagnostics[client_id] = {
          'delta_l2_norm': tree_util.tree_l2_norm(delta_params)
      }
    mean_delta_params = tree_util.tree_inverse_weight(delta_params_sum,
                                                      num_examples_sum)
    server_state = server_update(server_state, mean_delta_params)
    return server_state, client_diagnostics

  def server_update(server_state, mean_delta_params):
    opt_state, params = server_optimizer.apply(mean_delta_params,
                                               server_state.opt_state,
                                               server_state.params)
    return ServerState(params, opt_state)

  return federated_algorithm.FederatedAlgorithm(init, apply)
